/*
    Copyright (C) 2008-2019, Millistream Market Data <support@millistream.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "common.h"
#include "templates.h"
#include "encode.h"
#include "crypt.h"

/* buf must be at least 5 bytes */
static int add_len (uint8_t *buf, uint32_t len)
{
	uint8_t *p = buf;
	uint32_t tmp = len;
	uint32_t offset = 0;
	int d;
	int c = 0;

	do {
		++c;
		offset = offset * 128 + 1;
		tmp = (tmp - 1) / 128;
	} while (tmp != 0);

	len -= offset;
	d = c - 1;

	do {
		p[--c] = (uint8_t) (len % 128);
		len /= 128;
	} while (c != 0);

	p += d;

	*p++ |= 128;

	return p - buf;
}

/* buf must be at least 11 bytes */
static int add_header (uint8_t *buf, uint16_t mref, uint64_t insref)
{
	uint8_t *p = buf;
	uint8_t *attr = buf;

	if (mref <= 256) {
		*p++ = 0;
		*p++ = (uint8_t) mref;
	} else {
		*p++ = 128;
		*((uint16_t *)p) = cpu_to_le16 (mref);
		p += 2;
	}

	if (insref <= 8191) {
		*attr |= (uint8_t) (insref >> 8);
		*p++ = (uint8_t) (insref & __UINT64_C (0x00000000000000FF));
	} else if (insref <= 2105343) { 
		insref -= 8192;
		*attr |= 32 | (uint8_t) (insref >> 16);
		*((uint16_t *)p) = cpu_to_le16 ((uint16_t)(insref & __UINT64_C (0x000000000000FFFF)));
		p += 2;
	} else if (insref <= __UINT64_C (137441058815)) {
		insref -= 2105344;
		*attr |= 64 | (uint8_t) (insref >> 32);
		*((uint32_t *)p) = cpu_to_le32 ((uint32_t)(insref & __UINT64_C (0x00000000FFFFFFFF)));
		p += 4;
	} else {
		*attr |= 96;
		*((uint64_t *)p) = cpu_to_le64 (insref);
		p += 8;
	}

	return p - buf;
}

LIBMDF_API mdf_t mdf_create ()
{
	struct mdf_s *handle;
#if defined (HAVE_SIGACTION) && defined (SIGPIPE)
	struct sigaction act;
#endif
#if defined _WIN32 || defined _WIN64
	WSADATA	wsaData;

	if (WSAStartup (MAKEWORD (2,2), &wsaData) != 0)
		return NULL;
#endif
	handle = (struct mdf_s *) calloc (1, sizeof *handle);

	if (handle == NULL) {
#if defined _WIN32 || defined _WIN64
		WSACleanup ();
#endif
		return NULL;
	}

	handle->data_size = 10240;
	handle->data = (uint8_t *) malloc (handle->data_size);
	handle->timediff = 0;
	handle->vlen = 1024;
	handle->value = (uint8_t *) malloc (handle->vlen);
	handle->fstatus = 2;
	handle->error = MDF_ERR_NO_ERROR;
	handle->fd = INVALID_SOCKET;
	handle->connect_timeout = 5;
	handle->heartbeat_interval = 30;
	handle->max_missed_heartbeats = 2;
	handle->crypto.cstate = CSTATE_INVALID;

	if (handle->data == NULL || handle->value == NULL) {
#if defined _WIN32 || defined _WIN64
		WSACleanup ();
#endif
		free (handle->data);
		handle->data = NULL;

		free (handle->value);
		handle->value = NULL;

		free (handle);
		return NULL;
	}

	templates_initialize (&handle->templates, &handle->templates_num);

#if defined (HAVE_SIGACTION) && defined (SIGPIPE)
	memset (&act, 0, sizeof act);
	act.sa_handler = SIG_IGN;
	sigaction (SIGPIPE, &act, NULL);
#elif defined (HAVE_SIGNAL) && defined (SIGPIPE)
	signal (SIGPIPE, SIG_IGN);
#endif

	ENGINE_load_builtin_engines ();
	ENGINE_register_all_complete ();

	return handle;
}

LIBMDF_API void mdf_destroy (mdf_t handle)
{
	int i;

	if (handle == NULL)
		return;

	mdf_disconnect (handle);

	free (handle->data);
	handle->data = NULL;

	free (handle->value);
	handle->value = NULL;

	free (handle->connected_host);
	handle->connected_host = NULL;

	free (handle->connected_ip);
	handle->connected_ip = NULL;

	free (handle->bind_address);
	handle->bind_address = NULL;

	for (i = 0; i < handle->templates_num; i++) {
		free (handle->templates[i].tags);
		free (handle->templates[i].fields);
	}

	free (handle->templates);
	handle->templates = NULL;

	free (handle->sendfields);
	handle->sendfields = NULL;

	free (handle->sendtags);
	handle->sendtags = NULL;

	free (handle->sendbuf);
	handle->sendbuf = NULL;

	free (handle);

#if defined _WIN32 || defined _WIN64
	WSACleanup ();
#endif
}

static void call_status_callback (mdf_t handle, MDF_CONN_STATUS status, const char *host, const char *ip)
{
	if (handle->callbacks.func_status == NULL)
		return;

	handle->callbacks.func_status (handle->callbacks.udata_status, status, host, ip);
}

static void disconnect (mdf_t handle)
{
	if (handle->crypto.master_secret != NULL) {
		memset (handle->crypto.master_secret, 0x00, handle->crypto.digest_length);
		free (handle->crypto.master_secret);
		handle->crypto.master_secret = NULL;
	}

	if (handle->crypto.cipher_encrypt_ctx != NULL) {
		EVP_CIPHER_CTX_free (handle->crypto.cipher_encrypt_ctx);
		handle->crypto.cipher_encrypt_ctx = NULL;
	}

	if (handle->crypto.cipher_decrypt_ctx != NULL) {
		EVP_CIPHER_CTX_free (handle->crypto.cipher_decrypt_ctx);
		handle->crypto.cipher_decrypt_ctx = NULL;
	}

	if (handle->crypto.digest_encrypt_ctx != NULL) {
		EVP_MD_CTX_free (handle->crypto.digest_encrypt_ctx);
		handle->crypto.digest_encrypt_ctx = NULL;
	}

	if (handle->crypto.digest_decrypt_ctx != NULL) {
		EVP_MD_CTX_free (handle->crypto.digest_decrypt_ctx);
		handle->crypto.digest_decrypt_ctx = NULL;
	}

	handle->data_pos = 0;
	handle->data_used = 0;
	handle->fstatus = 2;
	handle->error = MDF_ERR_NO_ERROR;

#if defined _WIN32 || defined _WIN64
	closesocket (handle->fd);
#else
	close (handle->fd);
#endif
	handle->fd = INVALID_SOCKET;

	call_status_callback (handle, MDF_STATUS_DISCONNECTED, handle->connected_host, handle->connected_ip);
}

static int read_int_msg (mdf_t handle, uint8_t *buf, ssize_t * restrict len, int * restrict pos, const size_t maxlen)
{
	size_t offset = 0;
	size_t toread = 0;

	do {
		ssize_t ret;
		fd_set fdset;
		struct timeval tv;

		FD_ZERO (&fdset);
		FD_SET (handle->fd, &fdset);

		tv.tv_sec = 10;
		tv.tv_usec = 0;

		ret = select (handle->fd + 1, &fdset, NULL, NULL, &tv);

		if (ret != 1)
			return -1;

		ret = recv (handle->fd, (char *) (buf + offset), maxlen - offset, 0);

		if (ret < 1)
			return -1;

		handle->bytes_read += ret;
		offset += ret;

		if (offset < 5)
			continue;

		if (toread == 0) {
			int i = 0;

			do {
				toread *= 128;
				toread += (buf[i] & 127) + 1;
			} while ((buf[i++] & 128) == 0 && i < 5);

			*len = toread + i;
			*pos = i;

			toread += i;

			if (toread > maxlen)
				return -1;

			toread -= offset;
		} else
			toread -= ret;
	} while (toread != 0);

	return 0;
}

static int send_greeting (mdf_t handle)
{
	uint8_t *Nb = NULL;
	uint8_t *master_key;
	BIGNUM *X = NULL;
	RSA *rsa_master = NULL;
	RSA *rsa_server = NULL;
	EVP_MD_CTX *digest2_ctx = NULL;
#if OPENSSL_VERSION_NUMBER >= 0x10100000L
	BIGNUM *bn1 = NULL;
	BIGNUM *bn2 = NULL;
#endif

	/* As part of setting up the connection with the server, the client
	 * must initiate a crypto handshake with the server. The first
	 * message from the client will contain the mdf version (for book-
	 * keeping), a random nonce (Na), the minimum number of bytes that
	 * the client accepts for the DH p parameter, and a list of the
	 * digests and encryptions that we support. */
	ssize_t len = sizeof (MDF_VERSION) - 1;

	if (len > 255) {
		disconnect (handle);
		return 0;
	}

	ssize_t hostlen = sizeof (MDF_BUILD_HOST) - 1;

	if (hostlen > 255) {
		disconnect (handle);
		return 0;
	}

	/* 6 digests + 3 ciphers == 9 * sizeof (uint32_t) in the size below*/
	uint8_t *msg1 = malloc (5 + 11 + sizeof (uint64_t) + sizeof (uint8_t) + len + sizeof (uint32_t) + EVP_MAX_MD_SIZE + sizeof (uint32_t) + (9 * sizeof (uint32_t)) + sizeof (uint8_t) + hostlen);
	uint8_t *Na = malloc (EVP_MAX_MD_SIZE);

	int used = add_header (msg1 + 5, 0, 0) + 5;

	*((uint64_t *)(msg1 + used)) = cpu_to_le64 (handle->template_version);
	used += sizeof (uint64_t);

	msg1[used++] = (uint8_t) len;

	memcpy (msg1 + used, MDF_VERSION, len);
	used += len;

	/* we send the largest Na that we can produce */
	if (RAND_bytes (Na, EVP_MAX_MD_SIZE) != 1) {
		free (Na);
		free (msg1);
		disconnect (handle);
		return 0;
	}

	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (EVP_MAX_MD_SIZE);
	used += sizeof (uint32_t);
	memcpy (msg1 + used, Na, EVP_MAX_MD_SIZE);
	used += EVP_MAX_MD_SIZE;

	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (MIN_DH_LENGTH);
	used += sizeof (uint32_t);

	/* we add the list of our supported digest algorithms */
	int offset = used;
	used += sizeof (uint32_t);
	uint32_t tmp_u32 = 0;

#if defined HAVE_OPENSSL_MD5_H && !defined OPENSSL_NO_MD5
	tmp_u32++;
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (0);
	used += sizeof (uint32_t);
#endif
#if defined HAVE_OPENSSL_SHA_H && !defined OPENSSL_NO_SHA1
	tmp_u32++;
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (1);
	used += sizeof (uint32_t);
#endif
#if defined HAVE_OPENSSL_SHA_H && !defined OPENSSL_NO_SHA256
	tmp_u32++;
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (2);
	used += sizeof (uint32_t);
#endif
#if defined HAVE_OPENSSL_RIPEMD_H && !defined OPENSSL_NO_RIPEMD
	tmp_u32++;
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (3);
	used += sizeof (uint32_t);
#endif
#if defined HAVE_OPENSSL_SHA_H && !defined OPENSSL_NO_SHA512
	tmp_u32++;
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (4);
	used += sizeof (uint32_t);
#endif
#if defined HAVE_OPENSSL_WHRLPOOL_H && !defined OPENSSL_NO_WHIRLPOOL
	tmp_u32++;
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (5);
	used += sizeof (uint32_t);
#endif
	if (tmp_u32 == 0) {
		free (Na);
		free (msg1);
		disconnect (handle);
		return 0;
	}

	*((uint32_t *)(msg1 + offset)) = cpu_to_le32 (tmp_u32);

	/* we add the list of supported encryption methods */
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (3); used += sizeof (uint32_t); /* we add 3 ciphers */
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (0); used += sizeof (uint32_t);  /* AES 128-CTR */
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (1); used += sizeof (uint32_t);  /* AES 192-CTR */
	*((uint32_t *)(msg1 + used)) = cpu_to_le32 (2); used += sizeof (uint32_t);  /* AES 256-CTR */

	/* add the MDF_BUILD_HOST info */
	msg1[used++] = (uint8_t) hostlen;
	memcpy (msg1 + used, MDF_BUILD_HOST, hostlen);
	used += hostlen;

	int msg1_len = used - 5;

	offset = 5 - add_len (msg1, msg1_len);

	if (offset != 0) {
		memmove (msg1 + offset, msg1, 5 - offset);
		used -= offset;
	}

	len = send (handle->fd, (void *)(msg1 + offset), used, 0);

	if (len != used) {
		free (Na);
		free (msg1);
		disconnect (handle);
		return 0;
	}

	handle->bytes_written += len;

	/* We should now receive the first message from the server. This
	 * message will contain the public DH parameters (g, p), the
	 * server public DH key (X), a random nonce (Nb), the server
	 * public RSA key (e,n) and a signature from the "master" to
	 * authenticate the public RSA key, the hostname etc.
	 * Last in the message is a RSA signature created with the 
	 * server key of all bytes exchanged so far (by both parties).
	 *
	 * Since the signature covers the random nonce (Na) that the
	 * client generated, the client now knows two things:
	 *	1. The server has seen the first message from the client,
	 *	   so there was no man-in-the-middle fiddling with the
	 *	   message trying to impersonate as the client.
	 *
	 *	2. This message does indeed come from the server, it
	 *	   carries the servers signature and it is not part in
	 *	   any reply attack since Na was randomly chosen.
	*/
	uint8_t *buf = malloc (201728);

	if (read_int_msg (handle, buf, &len, &offset, 201728) == -1) {
		free (Na);
		free (buf);
		free (msg1);
		disconnect (handle);
		return 0;
	}

	int msg2_len = len - offset;

	EVP_MD_CTX *digest_ctx = NULL;
	DH *dh = DH_new ();

	if (len < 100)
		goto exit_in_error;

	const int msg2_offset = offset;

	if (buf[offset] != 0x00 || buf[offset + 1] != 0x00 || buf[offset + 2] != 0x03)
		goto exit_in_error;

	offset += 3;

	/* do we support the suggested encryption method? */
	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	const EVP_CIPHER *cipher;

	if (used == 0)
		cipher = EVP_aes_128_ctr();
	else if (used == 1)
		cipher = EVP_aes_192_ctr();
	else if (used == 2)
		cipher = EVP_aes_256_ctr();
	else
		goto exit_in_error;

	handle->crypto.cipher_encrypt_ctx = EVP_CIPHER_CTX_new ();
	EVP_EncryptInit_ex (handle->crypto.cipher_encrypt_ctx, cipher, NULL, NULL, NULL);

	handle->crypto.cipher_decrypt_ctx = EVP_CIPHER_CTX_new ();
	EVP_DecryptInit_ex (handle->crypto.cipher_decrypt_ctx, cipher, NULL, NULL, NULL);

	/* do we support the suggested digest method? */
	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	handle->crypto.digest = NULL;

#if defined HAVE_OPENSSL_MD5_H && !defined OPENSSL_NO_MD5
	if (used == 0) handle->crypto.digest = EVP_md5();
#endif
#if defined HAVE_OPENSSL_SHA_H && !defined OPENSSL_NO_SHA1
	if (used == 1) handle->crypto.digest = EVP_sha1();
#endif
#if defined HAVE_OPENSSL_SHA_H && !defined OPENSSL_NO_SHA256
	if (used == 2) handle->crypto.digest = EVP_sha256();
#endif
#if defined HAVE_OPENSSL_RIPEMD_H && !defined OPENSSL_NO_RIPEMD
	if (used == 3) handle->crypto.digest = EVP_ripemd160();
#endif
#if defined HAVE_OPENSSL_SHA_H && !defined OPENSSL_NO_SHA512
	if (used == 4) handle->crypto.digest = EVP_sha512();
#endif
#if defined HAVE_OPENSSL_WHRLPOOL_H && !defined OPENSSL_NO_WHIRLPOOL
	if (used == 5) handle->crypto.digest = EVP_whirlpool();
#endif
	if (handle->crypto.digest == NULL)
		goto exit_in_error;

	digest_ctx = EVP_MD_CTX_create ();
	EVP_DigestInit_ex (digest_ctx, handle->crypto.digest, NULL);

	handle->crypto.digest_length = EVP_MD_CTX_size (digest_ctx);

	/* since we now know the digest method of choice we can calculate what the
	 * digest of our first message should have been! */
	tmp_u32 = cpu_to_le32 (1); /* message no 1 */
	EVP_DigestUpdate (digest_ctx, &tmp_u32, sizeof tmp_u32);

	tmp_u32 = cpu_to_le32 (msg1_len);
	EVP_DigestUpdate (digest_ctx, &tmp_u32, sizeof tmp_u32);
	EVP_DigestUpdate (digest_ctx, msg1 + 5, msg1_len);

	free (msg1);
	msg1 = NULL;

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	/* do we have room for g and p-len? */
	if (offset + used + (int)sizeof (uint32_t) > len)
		goto exit_in_error;

#if OPENSSL_VERSION_NUMBER < 0x10100000L
	dh->g = BN_bin2bn (buf + offset, used, NULL);
#else
	bn1 = BN_bin2bn (buf + offset, used, NULL);
#endif
	offset += used;

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	/* do we have room for p and X-len? */
	if (offset + used + (int)sizeof (uint32_t) > len)
		goto exit_in_error;

	/* make sure that p is within limits */
	if (used < MIN_DH_LENGTH || used > MAX_DH_LENGTH)
		goto exit_in_error;

#if OPENSSL_VERSION_NUMBER < 0x10100000L
	dh->p = BN_bin2bn (buf + offset, used, NULL);
#else
	bn2 = BN_bin2bn (buf + offset, used, NULL);

	if (DH_set0_pqg (dh, bn2, NULL, bn1) == 0)
		goto exit_in_error;

	/* clear these so that we don't do a double free on exit */
	bn1 = NULL;
	bn2 = NULL;
#endif

	offset += used;

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	/* do we have room for X and Nb-len? */
	if (offset + used + (int)sizeof (uint32_t) > len)
		goto exit_in_error;

	X = BN_bin2bn (buf + offset, used, NULL);
	offset += used;

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	/* do we have room for Nb and master siglen? */
	if (offset + used + (int)sizeof (uint32_t) > len || used != handle->crypto.digest_length)
		goto exit_in_error;

	Nb = malloc (used);
	memcpy (Nb, buf + offset, used);
	offset += used;

	rsa_master = mdf_int_load_rsapublickey ();

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));

	const int rsa_size = RSA_size (rsa_master);

	/* do we have room for the master signature block? */
	if (offset + used + 1 > len || used <= rsa_size)
		goto exit_in_error;

	/* verify the master signature */
	unsigned char md[EVP_MAX_MD_SIZE];

	digest2_ctx = EVP_MD_CTX_create ();
	EVP_DigestInit_ex (digest2_ctx, handle->crypto.digest, NULL);
	EVP_DigestUpdate (digest2_ctx, buf + offset, used - rsa_size);
	EVP_DigestFinal_ex (digest2_ctx, md, NULL);

	if (RSA_verify (EVP_MD_CTX_type (digest2_ctx), md, handle->crypto.digest_length, buf + offset + (used - rsa_size), rsa_size, rsa_master) == 0)
		goto exit_in_error;

	offset += sizeof (uint32_t);

	/* verify the hostname from the master signature */
	used = 0;
	for (tmp_u32 = buf[offset++]; tmp_u32 != 0; tmp_u32--) {
		if (offset + buf[offset] + 1 > len)
			goto exit_in_error;

		if (buf[offset] == strlen (handle->connected_host) && memcmp (buf + offset + 1, handle->connected_host, buf[offset]) == 0)
			used = 1;
		else if (buf[offset] == strlen (handle->connected_ip) && memcmp (buf + offset + 1, handle->connected_ip, buf[offset]) == 0)
			used = 1;

		offset += buf[offset] + 1;
	}

	if (used == 0)
		goto exit_in_error;

	/* do we have room for the expiry time and e-len? */
	if (offset + (int)(sizeof (int64_t) + sizeof (uint32_t)) > len)
		goto exit_in_error;

	/* has the signature expired? */
	const int64_t expiry = le64_to_cpu (*((int64_t *)(buf + offset)));

	if (time (NULL) >= expiry)
		goto exit_in_error;

	offset += sizeof (int64_t);

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	/* do we have room for e, n-len and is e of the appropiate size? */
	if (offset + used + (int)sizeof (uint32_t) > len || used > OPENSSL_RSA_MAX_PUBEXP_BITS * 8 || used < 1)
		goto exit_in_error;

	rsa_server = RSA_new ();

#if OPENSSL_VERSION_NUMBER < 0x10100000L
	rsa_server->e = BN_bin2bn (buf + offset, used, NULL);
#else
	bn1 = BN_bin2bn (buf + offset, used, NULL);
#endif
	offset += used;

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	/* do we have room for n, signature-len and is n of the appropiate size? */
	if (offset + used + (int)sizeof (uint32_t) > len || used > OPENSSL_RSA_MAX_MODULUS_BITS * 8 || used < 256)
		goto exit_in_error;

#if OPENSSL_VERSION_NUMBER < 0x10100000L
	rsa_server->n = BN_bin2bn (buf + offset, used, NULL);
#else
	bn2 = BN_bin2bn (buf + offset, used, NULL);

	if (RSA_set0_key (rsa_server, bn2, bn1, NULL) == 0)
		goto exit_in_error;

	/* clear these so that we don't do a double free on exit */
	bn1 = NULL;
	bn2 = NULL;
#endif
	offset += used;

	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	/* do we have room for the signature and the sig-len, and is the signature of the correct size? */
	if (offset + used + (int)sizeof (uint32_t) > len || used != rsa_size)
		goto exit_in_error;

	/* ok, come this far we have now verified that we have got a valid RSA key from the server.
	 * We know that it hasn't expired, that it's hostname matches that to which we are connected to,
	 * and that the master signature was valid. */

	offset += used;
	RSA_free (rsa_master);
	rsa_master = NULL;

	/* let's look at the server's signature */
	used = le32_to_cpu (*((uint32_t *)(buf + offset)));
	offset += sizeof (uint32_t);

	tmp_u32 = cpu_to_le32 (2); /* message no 2 */
	EVP_DigestUpdate (digest_ctx, &tmp_u32, sizeof tmp_u32);

	tmp_u32 = cpu_to_le32 (msg2_len);
	EVP_DigestUpdate (digest_ctx, &tmp_u32, sizeof tmp_u32);
	EVP_DigestUpdate (digest_ctx, buf + msg2_offset, offset - msg2_offset);

	/* do we have room for the signature, and is the signature the correct size? */
	if (offset + used > len || used != rsa_size)
		goto exit_in_error;

	/* we make a copy of the CTX since we need to send an AUTH for the
	 * client as well and EVP_DigestFinal_ex() will destroy the context */
	EVP_MD_CTX_copy (digest2_ctx, digest_ctx);
	EVP_DigestFinal_ex (digest_ctx, md, NULL);

	if (RSA_verify (EVP_MD_CTX_type (digest_ctx), md, handle->crypto.digest_length, buf + offset, used, rsa_server) == 0)
		goto exit_in_error;

	RSA_free (rsa_server);
	rsa_server = NULL;

	/* the signature is included in the client AUTH */
	EVP_DigestUpdate (digest2_ctx, buf + offset, used);

	/* we check that p is a safe prime and that g is a suitable generator,
	 * since we wouldn't want to be catched with our pants down. And also
	 * we do this _after_ we checked the RSA signatures so we have at least
	 * some guarantees that we haven't received values that will keep us up
	 * all night performing the check, or whatever nasty ugliness that could
	 * have happened. */
	if (DH_check (dh, &used) == 0)
		goto exit_in_error;

	if (DH_generate_key (dh) == 0)
		goto exit_in_error;

	master_key = malloc (DH_size (dh));

	used = DH_compute_key (master_key, X, dh);

	if (used == -1) {
		free (master_key);
		goto exit_in_error;
	}

	BN_free (X);
	X = NULL;

	handle->crypto.master_secret = malloc (handle->crypto.digest_length);

	EVP_DigestInit_ex (digest_ctx, handle->crypto.digest, NULL);
	EVP_DigestUpdate (digest_ctx, master_key, used);
	EVP_DigestFinal_ex (digest_ctx, handle->crypto.master_secret, NULL);

	memset (master_key, 0x00, used);
	free (master_key);

	EVP_MD_CTX_free (digest_ctx);
	digest_ctx = NULL;

	handle->crypto.digest_encrypt_ctx = EVP_MD_CTX_create ();
	handle->crypto.digest_decrypt_ctx = EVP_MD_CTX_create ();

	mdf_int_set_keys (&handle->crypto, ENCRYPT, Na, sizeof md);
	mdf_int_set_keys (&handle->crypto, DECRYPT, Nb, handle->crypto.digest_length);

	free (Na);
	Na = NULL;

	free (Nb);
	Nb = NULL;

	/* Now we construct the final packet in our crypto handshake,
	 * this message will contain the clients public DH key and the
	 * client AUTH encrypted with the client encryption key.
	 *
	 * This AUTH will not authenticate the client as such (the server
	 * does not know which client it is until the client sends the 
	 * username/password credentials, and the client lacks a RSA key.
	 *
	 * No, the reason for the AUTH is to complete the handshake and to
	 * make sure that the server has a secure channel open with "someone"
	 * and that this "someone" was the very same entity who sent the very
	 * first message in the handshake. This is not as bad as it sounds,
	 * this is exactly like secure web browsing (aka SSL/TLS).
	 *
	 * The client on the other had does know that it is communicating
	 * with the server due to the RSA signature, and this is what is
	 * really important since it is the client who will have to send
	 * its password over the wire, and we better make sure it's to no
	 * stranger. */
	used = 5;

	used += add_header (buf + used, 0, 4);

#if OPENSSL_VERSION_NUMBER < 0x10100000L
	len = BN_bn2bin (dh->pub_key, buf + used + sizeof (uint32_t));
#else
	const BIGNUM *dh_pub, *dh_priv;
	DH_get0_key (dh, &dh_pub, &dh_priv);
	len = BN_bn2bin (dh_pub, buf + used + sizeof (uint32_t));
#endif

	*((uint32_t *)(buf + used)) = cpu_to_le32 (len);
	used += sizeof (uint32_t) + len;

	DH_free (dh);
	dh = NULL;

	*((uint32_t *)(buf + used)) = cpu_to_le32 (handle->crypto.digest_length);
	used += sizeof (uint32_t);

	/* len is now as far as we have to calculate AUTH, we have to
	 * perform this little tap dance since we need the message length
	 * attached to be able to calculate the correct message digest */
	len = used;
	used += handle->crypto.digest_length;

	tmp_u32 = cpu_to_le32 (3); /* message no 3 */
	EVP_DigestUpdate (digest2_ctx, &tmp_u32, sizeof tmp_u32);

	tmp_u32 = cpu_to_le32 (used - 5);
	EVP_DigestUpdate (digest2_ctx, &tmp_u32, sizeof tmp_u32);
	EVP_DigestUpdate (digest2_ctx, buf + 5, len - 5);

	offset = 5 - add_len (buf, used - 5);

	if (offset != 0) {
		memmove (buf + offset, buf, 5 - offset);
		used -= offset;
	}

	EVP_DigestFinal_ex (digest2_ctx, buf + len, NULL);
	EVP_MD_CTX_free (digest2_ctx);
	digest2_ctx = NULL;

	int outlen;
	EVP_EncryptUpdate (handle->crypto.cipher_encrypt_ctx, buf + len, &outlen, buf + len, handle->crypto.digest_length);
	handle->crypto.encrypt_cnt++;

	len = send (handle->fd, (void *)(buf + offset), used, 0);

	if (len != used)
		goto exit_in_error;

	handle->bytes_written += len;
	handle->crypto.cstate = CSTATE_ENCRYPT | CSTATE_DECRYPT;
	free (buf);
	return 1;

exit_in_error:
	disconnect (handle);

	if (rsa_master != NULL)
		RSA_free (rsa_master);

	if (rsa_server != NULL)
		RSA_free (rsa_server);

	if (dh != NULL)
		DH_free (dh);

	if (X != NULL)
		BN_free (X);

	if (digest_ctx != NULL)
		EVP_MD_CTX_free (digest_ctx);

	if (digest2_ctx != NULL)
		EVP_MD_CTX_free (digest2_ctx);

#if OPENSSL_VERSION_NUMBER >= 0x10100000L
	if (bn1 != NULL)
		BN_free (bn1);

	if (bn2 != NULL)
		BN_free (bn2);
#endif

	free (Na);
	free (Nb);
	free (buf);
	free (msg1);
	return 0;
}

void mdf_int_send_rekey (mdf_t handle)
{
	ssize_t written;
	int len;
	uint8_t buf[5 + 11 + sizeof (uint32_t) + EVP_MAX_MD_SIZE + EVP_MAX_MD_SIZE];
	uint8_t *p = buf;
	uint8_t *Na = NULL;
	uint8_t *first_pos;

	/* we don't support rekeyings on non encryption tunnels */
	if ((handle->crypto.cstate & CSTATE_ENCRYPT) != CSTATE_ENCRYPT)
		return;

	if (handle->no_encryption == 1)
		len = 3 + sizeof (uint32_t);
	else
		len = 3 + sizeof (uint32_t) + handle->crypto.digest_length;

	p += add_len (p, len);
	first_pos = p;

	p += add_header (p, 0, 5);

	if (handle->no_encryption == 1) {
		memset (p, 0, sizeof (uint32_t));
		p += sizeof (uint32_t);
	} else {
		*((uint32_t *)(p)) = cpu_to_le32 (handle->crypto.digest_length);
		p += sizeof (uint32_t);

		Na = malloc (handle->crypto.digest_length);

		if (RAND_bytes (Na, handle->crypto.digest_length) != 1) {
			free (Na);
			return;
		}

		memcpy (p, Na, handle->crypto.digest_length);
		p += handle->crypto.digest_length;
	}

	const uint32_t tmp_u32 = cpu_to_le32 ((p - buf) + handle->crypto.digest_length);
	const uint64_t tmp_u64 = cpu_to_le64 (handle->crypto.encrypt_cnt);

	EVP_DigestSignInit (handle->crypto.digest_encrypt_ctx, NULL, handle->crypto.digest, NULL, NULL);
	EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u64, sizeof tmp_u64);
	EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u32, sizeof tmp_u32);
	EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, buf, p - buf);

	size_t siglen = EVP_MAX_MD_SIZE;
	EVP_DigestSignFinal (handle->crypto.digest_encrypt_ctx, p, &siglen);
	p += siglen;

	int outlen;
	EVP_EncryptUpdate (handle->crypto.cipher_encrypt_ctx, first_pos, &outlen, first_pos, p - first_pos);
	handle->crypto.encrypt_cnt++;

	written = send (handle->fd, (void *)buf, p - buf, 0);

	if (written == -1) {
		free (Na);
		disconnect (handle);
		return;
	}

	handle->bytes_written += written;

	if (handle->no_encryption == 1) {
		handle->crypto.cstate &= ~CSTATE_ENCRYPT;

		EVP_CIPHER_CTX_free (handle->crypto.cipher_encrypt_ctx);
		handle->crypto.cipher_encrypt_ctx = NULL;

		EVP_MD_CTX_free (handle->crypto.digest_encrypt_ctx);
		handle->crypto.digest_encrypt_ctx = NULL;

		return;
	}

	mdf_int_set_keys (&handle->crypto, ENCRYPT, Na, handle->crypto.digest_length);
	free (Na);
}

static void should_we_rekey (mdf_t handle)
{
	if ((handle->crypto.cstate & CSTATE_ENCRYPT) != CSTATE_ENCRYPT)
		return;

	if (handle->crypto.encrypt_cnt >= MAX_ENCRYPTED_MESSAGES_PER_KEY)
		mdf_int_send_rekey (handle);
}

int mdf_int_send_heartbeat (mdf_t handle)
{
	ssize_t written;
	uint8_t buf[12 + EVP_MAX_MD_SIZE];
	uint8_t *p = buf;
	uint8_t *first_pos;

	p += add_len (p, 11);
	first_pos = p;

	p += add_header (p, 0, 2);

	(*(uint64_t *)p) = cpu_to_le64 (time (NULL));
	p += 8;

	if ((handle->crypto.cstate & CSTATE_ENCRYPT) == CSTATE_ENCRYPT) {
		const uint32_t tmp_u32 = cpu_to_le32 ((p - buf) + handle->crypto.digest_length);
		const uint64_t tmp_u64 = cpu_to_le64 (handle->crypto.encrypt_cnt);

		EVP_DigestSignInit (handle->crypto.digest_encrypt_ctx, NULL, handle->crypto.digest, NULL, NULL);
		EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u64, sizeof tmp_u64);
		EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u32, sizeof tmp_u32);
		EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, buf, p - buf);

		size_t siglen = EVP_MAX_MD_SIZE;
		EVP_DigestSignFinal (handle->crypto.digest_encrypt_ctx, p, &siglen);
		p += siglen;

		int outlen;
		EVP_EncryptUpdate (handle->crypto.cipher_encrypt_ctx, first_pos, &outlen, first_pos, p - first_pos);
		handle->crypto.encrypt_cnt++;
	}

	written = send (handle->fd, (void *)buf, p - buf, 0);

	if (written == -1) {
		disconnect (handle);
		handle->error = MDF_ERR_DISCONNECTED;
		return -1;
	}

	handle->bytes_written += written;
	should_we_rekey (handle);
	return 0;
}

static int request_heartbeat (mdf_t handle)
{
	ssize_t written;
	uint8_t buf[5 + 11 + EVP_MAX_MD_SIZE];
	uint8_t *p = buf;
	uint8_t *first_pos;

	p += add_len (p, 3);
	first_pos = p;

	p += add_header (p, 0, 1);

	if ((handle->crypto.cstate & CSTATE_ENCRYPT) == CSTATE_ENCRYPT) {
		const uint32_t tmp_u32 = cpu_to_le32 ((p - buf) + handle->crypto.digest_length);
		const uint64_t tmp_u64 = cpu_to_le64 (handle->crypto.encrypt_cnt);

		EVP_DigestSignInit (handle->crypto.digest_encrypt_ctx, NULL, handle->crypto.digest, NULL, NULL);
		EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u64, sizeof tmp_u64);
		EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u32, sizeof tmp_u32);
		EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, buf, p - buf);

		size_t siglen = EVP_MAX_MD_SIZE;
		EVP_DigestSignFinal (handle->crypto.digest_encrypt_ctx, p, &siglen);
		p += handle->crypto.digest_length;

		int outlen;
		EVP_EncryptUpdate (handle->crypto.cipher_encrypt_ctx, first_pos, &outlen, first_pos, p - first_pos);
		handle->crypto.encrypt_cnt++;
	}

	written = send (handle->fd, (void *)buf, p - buf, 0);

	if (written == -1) {
		disconnect (handle);
		handle->error = MDF_ERR_DISCONNECTED;
		return -1;
	}

	handle->bytes_written += written;
	should_we_rekey (handle);
	return 0;
}

#if defined _WIN32 || defined _WIN64
static int is_connected (SOCKET fd, fd_set *rdevents, fd_set *wrevents, fd_set *exevents)
{
	WSASetLastError (0);

	if (!FD_ISSET (fd, rdevents) && !FD_ISSET (fd, wrevents))
		return 0;

	if (FD_ISSET (fd, exevents))
		return 0;

	return 1;
}
#else
static int is_connected (int fd, fd_set *rdevents, fd_set *wrevents, fd_set *exevents)
{
	int err;
	socklen_t len = sizeof err;

	(void)exevents;

	errno = 0;

	if (!FD_ISSET (fd, rdevents) && !FD_ISSET (fd, wrevents))
		return 0;

	if (getsockopt (fd, SOL_SOCKET, SO_ERROR, &err, &len) < 0)
		return 0;

	errno = err;

	return err == 0 ? 1 : 0;
}
#endif

static int try_connect (mdf_t handle, struct addrinfo *addr)
{
	fd_set rdevents;
	fd_set wrevents;
	fd_set exevents;
	struct timeval tv;
	int ret;

#if defined _WIN32 || defined _WIN64
	unsigned long val = 1;
#else
	int val;
#endif

	handle->fd = socket (addr->ai_family, addr->ai_socktype | SOCK_CLOEXEC, addr->ai_protocol);

	if (handle->fd == INVALID_SOCKET)
		return 0;

	/* set non-blocking */
#if defined _WIN32 || defined _WIN64
	if (ioctlsocket (handle->fd, FIONBIO, &val) == SOCKET_ERROR) {
		closesocket (handle->fd);
		handle->fd = INVALID_SOCKET;
		return 0;
	}
#else
	val = fcntl (handle->fd, F_GETFL, 0);

	if (val == -1) {
		close (handle->fd);
		handle->fd = INVALID_SOCKET;
		return 0;
	}

	if (fcntl (handle->fd, F_SETFL, val | O_NONBLOCK) == -1) {
		close (handle->fd);
		handle->fd = INVALID_SOCKET;
		return 0;
	}
#endif
	if (handle->bind_address != NULL) {
		struct addrinfo hint;
		struct addrinfo *res;

		memset (&hint, 0, sizeof hint);

		hint.ai_flags    = AI_ADDRCONFIG | AI_NUMERICSERV;
		hint.ai_family   = addr->ai_family;
		hint.ai_socktype = addr->ai_socktype;
		hint.ai_protocol = addr->ai_protocol;

		if (getaddrinfo (handle->bind_address, "0", &hint, &res) != 0) {
#if defined _WIN32 || defined _WIN64
			closesocket (handle->fd);
#else
			close (handle->fd);
#endif
			handle->fd = INVALID_SOCKET;
			return 0;
		}

		if (bind (handle->fd, res->ai_addr, res->ai_addrlen) != 0) {
			freeaddrinfo (res);
#if defined _WIN32 || defined _WIN64
			closesocket (handle->fd);
#else
			close (handle->fd);
#endif
			handle->fd = INVALID_SOCKET;
			return 0;
		}

		freeaddrinfo (res);
	}

	ret = connect (handle->fd, addr->ai_addr, addr->ai_addrlen);

#if defined _WIN32 || defined _WIN64
	if (ret == SOCKET_ERROR) {
		if (WSAGetLastError () != WSAEWOULDBLOCK) {
			closesocket (handle->fd);
			handle->fd = INVALID_SOCKET;
			return 0;
		}
#else
	if (ret == -1) {
		if (errno != EINPROGRESS) {
			close (handle->fd);
			handle->fd = INVALID_SOCKET;
			return 0;
		}
#endif
		FD_ZERO (&rdevents);
		FD_SET (handle->fd, &rdevents);

		wrevents = rdevents;
		exevents = rdevents;

		tv.tv_sec = handle->connect_timeout;
		tv.tv_usec = 0;

		if (select (handle->fd + 1, &rdevents, &wrevents, &exevents, &tv) < 1) {
#if defined _WIN32 || defined _WIN64
			closesocket (handle->fd);
#else
			close (handle->fd);
#endif
			handle->fd = INVALID_SOCKET;
			return 0;
		}

		if (is_connected (handle->fd, &rdevents, &wrevents, &exevents) == 0) {
#if defined _WIN32 || defined _WIN64
			closesocket (handle->fd);
#else
			close (handle->fd);
#endif
			handle->fd = INVALID_SOCKET;
			return 0;
		}
	}

	/* set socket back to blocking */
#if defined _WIN32 || defined _WIN64
	val = 0;
	ioctlsocket (handle->fd, FIONBIO, &val);
#else
	fcntl (handle->fd, F_SETFL, val);
#endif

	return 1;
}

static int try_host (mdf_t handle, const char * restrict host, const char * restrict port)
{
	struct addrinfo hint;
	struct addrinfo *res;
	struct addrinfo *r;

	memset (&hint, 0, sizeof hint);

	hint.ai_flags    = AI_ADDRCONFIG | AI_NUMERICSERV;
	hint.ai_family   = AF_UNSPEC;
	hint.ai_socktype = SOCK_STREAM;
	hint.ai_protocol = IPPROTO_TCP;

	call_status_callback (handle, MDF_STATUS_LOOKUP, host, NULL);

	if (getaddrinfo (host, port, &hint, &res) != 0)
		return 0;

	for (r = res; r != NULL; r = r->ai_next) {
		char buf[50];
		char *ip = NULL;

		if (getnameinfo (r->ai_addr, r->ai_addrlen, buf, sizeof buf, NULL, 0, NI_NUMERICHOST) == 0)
			ip = buf;

		call_status_callback (handle, MDF_STATUS_CONNECTING, host, ip);

		if (try_connect (handle, r) == 1) {
			call_status_callback (handle, MDF_STATUS_CONNECTED, host, ip);
			freeaddrinfo (res);

			free (handle->connected_host);
			handle->connected_host = strdup (host);

			free (handle->connected_ip);
			handle->connected_ip = strdup (ip);
			return 1;
		}
	}

	freeaddrinfo (res);

	return 0;
}

#ifdef HAVE_SYS_UN_H
static int try_unix (mdf_t handle, const char *path)
{
	union { struct sockaddr_un un; struct sockaddr addr; } addr;
	int ret;

	if (strlen (path) + 1 > sizeof addr.un.sun_path)
		return 0;

	handle->fd = socket (AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);

	if (handle->fd == INVALID_SOCKET)
		return 0;

	memset (&addr.un, 0, sizeof addr.un);

	addr.un.sun_family = AF_UNIX;
	strcpy (addr.un.sun_path, path);

	ret = connect (handle->fd, &addr.addr, sizeof addr.un);

	if (ret == -1) {
		close (handle->fd);
		handle->fd = INVALID_SOCKET;
		return 0;
	}

	call_status_callback (handle, MDF_STATUS_CONNECTED, path, path);

	free (handle->connected_host);
	handle->connected_host = strdup (path);

	free (handle->connected_ip);
	handle->connected_ip = strdup (path);

	handle->crypto.cstate = 0; /* No need to encrypt data on AF_UNIX */
	return 1;
}
#endif

LIBMDF_API int mdf_connect (mdf_t handle, const char *servers)
{
	char *s, *p, *e, *port;

	if (handle == NULL)
		return 0;

	if (servers == NULL) {
		handle->error = MDF_ERR_ARGUMENT;
		return 0;
	}

	/* if we already are connected we must fail */
	if (handle->fd != INVALID_SOCKET) {
		handle->error = MDF_ERR_CONNECTED;
		return 0;
	}

	handle->crypto.cstate = CSTATE_INVALID;

	s = strdup (servers);

	/* parse the servers string */
	for (p = s; *p != '\0'; p = e) {
		while (isspace ((unsigned char) *p) != 0 || *p == ',')
			++p;

		/* IPv6 */
		if (*p == '[') {
			++p;

			port = strchr (p, ']');

			if (port == NULL) {
				free (s);
				handle->error = MDF_ERR_ARGUMENT;
				return 0;
			}

			*port++ = '\0';
#ifdef HAVE_SYS_UN_H
		} else if (*p == '/') { /* Unix Domain Sockets */
			e = strchr (p, ',');

			if (e != NULL)
				*e++ = '\0';

			if (try_unix (handle, p) == 1) {
				free (s);
				handle->time_lastupdate = time (NULL);
				handle->heartbeats_sent = 0;
				return 1;
			}

			if (e == NULL)
				break;

			continue;
#endif
		} else { /* IPv4 or DNS name */
			port = strchr (p, ':');

			if (port == NULL) {
				free (s);
				handle->error = MDF_ERR_ARGUMENT;
				return 0;
			}
		}

		/* the format must be host:port */
		if (*port != ':') {
			free (s);
			handle->error = MDF_ERR_ARGUMENT;
			return 0;
		}

		*port++ = '\0';

		/* scan to the end of the port, yes this code does not perform
		 * all the checks that it could do, i.e we do not check if the port
		 * was written like host:1234newhost, but it is futile to try every
		 * possible error, we should just be happy that we at least do not
		 * crash ;) */
		for (e = port; isdigit ((unsigned char) *e) != 0; e++)
			;

		/* do not increase the end pointer if we are at the last byte in the string */
		if (*e != '\0')
			*e++ = '\0';

		/* we found one, try it! */
		if (try_host (handle, p, port) != 0) {
			setsockopt (handle->fd, SOL_TCP, TCP_NODELAY, (char *) &handle->tcp_nodelay, sizeof handle->tcp_nodelay);

			if (send_greeting (handle) == 1) {
				free (s);
				handle->time_lastupdate = time (NULL);
				handle->heartbeats_sent = 0;
				call_status_callback (handle, MDF_STATUS_READYTOLOGON, handle->connected_host, handle->connected_ip);
				return 1;
			}
		}
	}

	free (s);
	handle->error = MDF_ERR_CONNECT;
	return 0;
}

LIBMDF_API void mdf_disconnect (mdf_t handle)
{
	if (handle == NULL || handle->fd == INVALID_SOCKET)
		return;

	disconnect (handle);
}

LIBMDF_API int mdf_get_property (mdf_t handle, MDF_OPTION option, ...)
{
	va_list ap;

	if (handle == NULL)
		return 0;

	va_start (ap, option);

	switch (option) {
		case MDF_OPT_FD : {
#if defined _WIN32 || defined _WIN64
			SOCKET *value = va_arg (ap, SOCKET *);
#else
			int *value = va_arg (ap, int *);
#endif
			if (value == NULL)
				return 0;

			*value = handle->fd;
			return 1;
		}

		case MDF_OPT_ERROR : {
			MDF_ERROR *value = va_arg (ap, MDF_ERROR *);

			if (value == NULL)
				return 0;

			*value = handle->error;
			return 1;
		}

		case MDF_OPT_RCV_BYTES : {
			uint64_t *value = va_arg (ap, uint64_t *);

			if (value == NULL)
				return 0;

			*value = handle->bytes_read;
			return 1;
		}

		case MDF_OPT_SENT_BYTES : {
			uint64_t *value = va_arg (ap, uint64_t *);

			if (value == NULL)
				return 0;

			*value = handle->bytes_written;
			return 1;
		}

		case MDF_OPT_DATA_CALLBACK_FUNCTION : {
			mdf_data_callback *value = va_arg (ap, mdf_data_callback *);

			if (value == NULL)
				return 0;

			*value = handle->callbacks.func_data;
			return 1;
		}

		case MDF_OPT_DATA_CALLBACK_USERDATA : {
			char **value = va_arg (ap, char **);

			if (value == NULL)
				return 0;

			*value = (char *) handle->callbacks.udata_data;
			return 1;
		}

		case MDF_OPT_STATUS_CALLBACK_FUNCTION : {
			mdf_status_callback *value = va_arg (ap, mdf_status_callback *);

			if (value == NULL)
				return 0;

			*value = handle->callbacks.func_status;
			return 1;
		}

		case MDF_OPT_STATUS_CALLBACK_USERDATA : {
			char **value = va_arg (ap, char **);

			if (value == NULL)
				return 0;

			*value = (char *) handle->callbacks.udata_status;
			return 1;
		}

		case MDF_OPT_CONNECT_TIMEOUT : {
			int *value = va_arg (ap, int *);

			if (value == NULL)
				return 0;

			*value = handle->connect_timeout;
			return 1;
		}

		case MDF_OPT_HEARTBEAT_INTERVAL : {
			int *value = va_arg (ap, int *);

			if (value == NULL)
				return 0;

			*value = handle->heartbeat_interval;
			return 1;
		}

		case MDF_OPT_HEARTBEAT_MAX_MISSED : {
			int *value = va_arg (ap, int *);

			if (value == NULL)
				return 0;

			*value = handle->max_missed_heartbeats;
			return 1;
		}

		case MDF_OPT_TCP_NODELAY : {
			int *value = va_arg (ap, int *);

			if (value == NULL)
				return 0;

			*value = handle->tcp_nodelay;
			return 1;
		}

		case MDF_OPT_NO_ENCRYPTION : {
			int *value = va_arg (ap, int *);

			if (value == NULL)
				return 0;

			*value = handle->no_encryption;
			return 1;
		}

		case MDF_OPT_TIME_DIFFERENCE : {
			int *value = va_arg (ap, int *);

			if (value == NULL)
				return 0;

			*value = handle->timediff;
			return 1;
		}

		case MDF_OPT_BIND_ADDRESS : {
			char **value = va_arg (ap, char **);

			if (value == NULL)
				return 0;

			*value = (char *) handle->bind_address;
			return 1;
		}
	}

	va_end (ap);

	return 0;
}

LIBMDF_API int mdf_set_property (mdf_t handle, MDF_OPTION option, void *value)
{
	if (handle == NULL)
		return 0;

	switch (option) {
		case MDF_OPT_DATA_CALLBACK_FUNCTION :
			handle->callbacks.func_data = (mdf_data_callback) value;
			return 1;

		case MDF_OPT_DATA_CALLBACK_USERDATA :
			handle->callbacks.udata_data = value;
			return 1;

		case MDF_OPT_STATUS_CALLBACK_FUNCTION :
			handle->callbacks.func_status = (mdf_status_callback) value;
			return 1;

		case MDF_OPT_STATUS_CALLBACK_USERDATA :
			handle->callbacks.udata_status = value;
			return 1;

		case MDF_OPT_CONNECT_TIMEOUT : {
			int *val = (int *) value;

			if (*val < 1 || *val > 60)
				return 0;

			handle->connect_timeout = *val;
			return 1;
		}

		case MDF_OPT_RCV_BYTES :
			handle->bytes_read = *((uint64_t *) value);
			return 1;

		case MDF_OPT_SENT_BYTES :
			handle->bytes_written = *((uint64_t *) value);
			return 1;

		case MDF_OPT_ERROR :
			handle->error = *((MDF_ERROR *) value);
			return 1;

		case MDF_OPT_FD :
			return 0;

		case MDF_OPT_HEARTBEAT_INTERVAL : {
			int *val = (int *) value;

			if (*val < 1 || *val > 86400)
				return 0;

			handle->heartbeat_interval = *val;
			return 1;
		}

		case MDF_OPT_HEARTBEAT_MAX_MISSED : {
			int *val = (int *) value;

			if (*val < 1 || *val > 100)
				return 0;

			handle->max_missed_heartbeats = *val;
			return 1;
		}

		case MDF_OPT_TCP_NODELAY : {
			int *val = (int *) value;

			if (*val < 0 || *val > 1)
				return 0;

			handle->tcp_nodelay = *val;

			if (handle->fd != INVALID_SOCKET)
				setsockopt (handle->fd, SOL_TCP, TCP_NODELAY, (char *) &handle->tcp_nodelay, sizeof handle->tcp_nodelay);

			return 1;
		}

		case MDF_OPT_NO_ENCRYPTION : {
			int *val = (int *) value;

			if (*val < 0 || *val > 1)
				return 0;

			handle->no_encryption = *val;
			return 1;
		}

		case MDF_OPT_TIME_DIFFERENCE :
			return 0;

		case MDF_OPT_BIND_ADDRESS : {
			free (handle->bind_address);

			handle->bind_address = value == NULL ? NULL : strdup ((char *)value);
			return 1;
		}
	}

	return 0;
}

LIBMDF_API int mdf_consume (mdf_t handle, const int timeout)
{
	fd_set rdset;
	struct timeval tv;
	int ret;
	int flags = 0;
	int pos;
	size_t tlen;

	if (handle == NULL)
		return -1;

	if (handle->fd == INVALID_SOCKET) {
		handle->error = MDF_ERR_NOT_CONNECTED;
		return -1;
	}

	if ((handle->crypto.cstate & CSTATE_INVALID) == CSTATE_INVALID)
		return 0;

#ifdef MSG_DONTWAIT
	flags = MSG_DONTWAIT;

	/* if we are on a real POSIX system and timeout is zero
	 * we can skip the whole select() call saving us some cs */
	if (timeout != 0) {
#endif

	FD_ZERO (&rdset);
	FD_SET (handle->fd, &rdset);

	if (timeout < 0) {
		if (timeout < -999999) {
			tv.tv_sec = 1;
			tv.tv_usec = 0;
		} else {
			tv.tv_sec = 0;
			tv.tv_usec = -timeout;
		}
	} else if (timeout > 60 * 60) {
		tv.tv_sec = 60 * 60;
		tv.tv_usec = 0;
	} else {
		tv.tv_sec = timeout;
		tv.tv_usec = 0;
	}

	ret = select (handle->fd + 1, &rdset, NULL, NULL, &tv);

	if (ret == 0) {
		/* is it time to send a heartbeat */
		time_t now = time (NULL);

		if (difftime (now, handle->time_lastupdate) >= handle->heartbeat_interval) {
			if (handle->heartbeats_sent++ == handle->max_missed_heartbeats) {
				disconnect (handle);
				handle->error = MDF_ERR_CONNECTION_IDLE;
				return -1;
			}

			handle->time_lastupdate = now;
			return request_heartbeat (handle);
		}

		return 0;
	}

	if (ret < 0) {
#ifdef HAVE_ERRNO_H
		if (errno == EINTR)
			return 0;
#endif
		disconnect (handle);
		handle->error = MDF_ERR_DISCONNECTED;
		return -1;
	}
#ifdef MSG_DONTWAIT
	}
#endif

#ifdef MSG_DONTWAIT
try_read_again:
#endif
	ret = recv (handle->fd, (char *) (handle->data + handle->data_used), handle->data_size - handle->data_used, flags);

	if (ret == 0) {
		disconnect (handle);
		handle->error = MDF_ERR_DISCONNECTED;
		return -1;
	}

	if (ret < 0) {
#ifdef MSG_DONTWAIT
		if (errno == EAGAIN) {
			/* is it time to send a heartbeat */
			time_t now = time (NULL);

			if (difftime (now, handle->time_lastupdate) >= handle->heartbeat_interval) {
				if (handle->heartbeats_sent++ == handle->max_missed_heartbeats) {
					disconnect (handle);
					handle->error = MDF_ERR_CONNECTION_IDLE;
					return -1;
				}

				handle->time_lastupdate = now;
				return request_heartbeat (handle);
			}

			return 0;
		}

		if (errno == EINTR)
			goto try_read_again;
#endif
		disconnect (handle);
		handle->error = MDF_ERR_DISCONNECTED;
		return -1;
	}

	handle->time_lastupdate = time (NULL);
	handle->bytes_read += ret;
	handle->data_used += ret;
	handle->heartbeats_sent = 0;

	/* in case we get called before the client has decoded all the data in the buffer */
	if (handle->fstatus != 2) {
		if (handle->data_used - ret == handle->message_end)
			handle->oldval = handle->data[handle->message_end];

		if (handle->callbacks.func_data == NULL)
			return 1;

		do {
			handle->callbacks.func_data (handle->callbacks.udata_data, handle);
		} while (handle->fstatus != 2);

		return 0;
	}

	pos = 0;
	tlen = 0;

	do {
		tlen *= 128;
		tlen += (handle->data[handle->data_pos + pos] & 127) + 1;
	} while ((handle->data[handle->data_pos + pos++] & 128) == 0 && handle->data_pos + pos < handle->data_used);

	if (pos > 5) {
		disconnect (handle);
		handle->error = MDF_ERR_MSG_OOB;
		return -1;
	}

	/* we have not yet received a full header */
	if (pos == 0 || (handle->data[handle->data_pos + pos - 1] & 128) == 0) {
		if (handle->data_pos != 0) {
			handle->data_used -= handle->data_pos;
			memmove (handle->data, handle->data + handle->data_pos, handle->data_used);
			handle->data_pos = 0;
		}

		return 0;
	}

	handle->message_len = tlen + pos;
	handle->message_end = handle->data_pos + handle->message_len;
	handle->pos = pos;
	handle->fpos = handle->data_pos + pos;

	if ((handle->crypto.cstate & CSTATE_DECRYPT) == CSTATE_DECRYPT)
		handle->message_len += handle->crypto.digest_length;

	if (handle->data_pos + handle->message_len + 12 > handle->data_size) {
		void *tmp = realloc (handle->data, handle->data_pos + handle->message_len + 12);

		if (tmp == NULL) {
			disconnect (handle);
			handle->error = MDF_ERR_NO_MEM;
			return -1;
		}

		handle->data = (uint8_t *) tmp;
		handle->data_size = handle->data_pos + handle->message_len + 12;
	}

	if (handle->message_len + handle->data_pos > handle->data_used) {
		if (handle->data_pos != 0) {
			handle->data_used -= handle->data_pos;
			memmove (handle->data, handle->data + handle->data_pos, handle->data_used);
			handle->data_pos = 0;
		}

		return 0;
	}

	handle->fstatus = 1;

	if (handle->callbacks.func_data == NULL)
		return 1;

	do {
		handle->callbacks.func_data (handle->callbacks.udata_data, handle);
	} while (handle->fstatus != 2);

	return 0;
}

LIBMDF_API int mdf_message_send (mdf_t handle, mdf_message_t message)
{
	unsigned int i;
	struct field *field;

	if (handle == NULL || message == NULL)
		return 0;

	if (handle->fd == INVALID_SOCKET)
		return 0;

	if ((handle->crypto.cstate & CSTATE_INVALID) == CSTATE_INVALID)
		return 0;

	if (message->current_message == -1)
		return 0;

#ifdef TCP_CORK
	i = 1;
	setsockopt (handle->fd, IPPROTO_TCP, TCP_CORK, &i, sizeof (i));
#endif

	field = message->fields;

	for (i = 0; i < (unsigned int)(message->current_message + 1); i++) {
		uint8_t *p, *pmap_pos;
		unsigned int j;
		unsigned int tags_used = 0;
		unsigned int maxpos = 0;
		int hlen, offset;
		size_t len = 0;
		struct msg_template *template;
		ssize_t written;

		if (message->messages[i].fields == 0 && message->messages[i].mref != MDF_M_INSTRUMENTRESET && message->messages[i].mref != MDF_M_INSTRUMENTDELETE && message->messages[i].mref != MDF_M_ORDERBOOKFLUSH && message->messages[i].mref != MDF_M_LOGOFF)
			continue;

		if (message->messages[i].mref > 0 && message->messages[i].mref < handle->templates_num) {
			template = &(handle->templates[message->messages[i].mref]);

			if (template->tags_num > handle->sendfields_num) {
				free (handle->sendfields);
				handle->sendfields = malloc (sizeof *handle->sendfields * template->tags_num);
				handle->sendfields_num = template->tags_num;

				if (handle->sendfields == NULL) {
#ifdef TCP_CORK
					i = 0;
					setsockopt (handle->fd, IPPROTO_TCP, TCP_CORK, &i, sizeof (i));
#endif
					handle->sendfields_num = 0;
					handle->error = MDF_ERR_NO_MEM;
					return 0;
				}
			}

			memset (handle->sendfields, 0, sizeof *handle->sendfields * template->tags_num);
		} else {
			template = NULL;
		}

		/* create one list of the tags in positional order and one list of the
		 * tags that we must send as "tagged" */
		for (j = 0; j < (unsigned int)(message->messages[i].fields); j++) {
			/* perform binary search for the field position */
			int high = template == NULL ? -1 : template->tags_num - 1;
			int pos = -1;
			int low = 0;

			while (high >= low) {
				const int mid = (low + high) / 2;

				if (template->fields[mid].tag == field->tag) {
					pos = template->fields[mid].pos;
					break;
				}

				if (template->fields[mid].tag > field->tag)
					high = mid - 1;
				else
					low = mid + 1;
			}

			if (pos != -1) {
				if (handle->sendfields[pos] != NULL)
					len -= handle->sendfields[pos]->len;

				handle->sendfields[pos] = field;

				if (pos > (int)maxpos)
					maxpos = pos;
			} else {
				if (tags_used + 1 > handle->sendtags_num) {
					struct field **tmp = handle->sendtags;
					handle->sendtags = realloc (handle->sendtags, sizeof *handle->sendtags * (tags_used + 1));
					handle->sendtags_num = tags_used + 1;

					if (handle->sendtags == NULL) {
#ifdef TCP_CORK
						i = 0;
						setsockopt (handle->fd, IPPROTO_TCP, TCP_CORK, &i, sizeof (i));
#endif
						free (tmp);
						handle->sendtags_num = 0;
						handle->error = MDF_ERR_NO_MEM;
						return 0;
					}
				}

				handle->sendtags[tags_used++] = field;

				if (field->tag <= 255)
					len += 2;
				else if (field->tag <= 65535)
					len += 3;
				else
					len += 5;
			}

			len += field->len;
			field++;
		}

		/* the number of pmap:s that we will use */
		len += (maxpos / 8) + 1;

		if (len + 5 + 11 + handle->crypto.digest_length > (size_t) handle->sendbuf_size) {
			free (handle->sendbuf);
			handle->sendbuf = malloc (len + 5 + 11 + handle->crypto.digest_length);
			handle->sendbuf_size = len + 5 + 11 + handle->crypto.digest_length;

			if (handle->sendbuf == NULL) {
#ifdef TCP_CORK
				i = 0;
				setsockopt (handle->fd, IPPROTO_TCP, TCP_CORK, &i, sizeof (i));
#endif
				handle->sendbuf_size = 0;
				handle->error = MDF_ERR_NO_MEM;
				return 0;
			}
		}

		p = handle->sendbuf + 5;
		hlen = add_header (p, (uint16_t) (message->messages[i].mref + 1), message->messages[i].insref);
		len += hlen;
		p += hlen;

		hlen = add_len (handle->sendbuf, len);
		offset = 5 - hlen;

		if (offset != 0)
			memmove (handle->sendbuf + offset, handle->sendbuf, hlen);

		if (template != NULL) {
			int pmap = 0;

			for (j = 0; j < maxpos + 1; j++) {
				if (pmap == 0) {
					pmap = 128;
					pmap_pos = p;
					*p++ = 0;
				}

				if (handle->sendfields[j] != NULL) {
					*pmap_pos |= pmap;

					if (tags_used != 0) {
						unsigned int k;

						for (k = 0; k < tags_used; k++) {
							if (handle->sendtags[k]->tag <= 255) {
								*p++ = 251;
								*p++ = (uint8_t) handle->sendtags[k]->tag;
							} else if (handle->sendtags[k]->tag <= 65535) {
								*p++ = 252;
								*((uint16_t *) p) = cpu_to_le16 ((uint16_t) handle->sendtags[k]->tag);
								p += 2;
							} else {
								*p++ = 253;
								*((uint32_t *) p) = cpu_to_le32 ((uint32_t) handle->sendtags[k]->tag);
								p += 4;
							}

							memcpy (p, message->data + handle->sendtags[k]->offset, handle->sendtags[k]->len);
							p += handle->sendtags[k]->len;
						}

						tags_used = 0;
					}

					memcpy (p, message->data + handle->sendfields[j]->offset, handle->sendfields[j]->len);
					p += handle->sendfields[j]->len;
				}

				pmap /= 2;
			}
		} else if (tags_used == 0) {
			*p++ = 0;
		}

		/* apparently we had only tagged data today... */
		if (tags_used != 0) {
			if (template == NULL)
				*p++ = 128;
			else
				*pmap_pos = 128;

			for (j = 0; j < tags_used; j++) {
				if (handle->sendtags[j]->tag <= 255) {
					*p++ = 251;
					*p++ = (uint8_t) handle->sendtags[j]->tag;
				} else if (handle->sendtags[j]->tag <= 65535) {
					*p++ = 252;
					*((uint16_t *) p) = cpu_to_le16 ((uint16_t) handle->sendtags[j]->tag);
					p += 2;
				} else {
					*p++ = 253;
					*((uint32_t *) p) = cpu_to_le32 ((uint32_t) handle->sendtags[j]->tag);
					p += 4;
				}

				memcpy (p, message->data + handle->sendtags[j]->offset, handle->sendtags[j]->len);
				p += handle->sendtags[j]->len;
			}
		}

		if ((handle->crypto.cstate & CSTATE_ENCRYPT) == CSTATE_ENCRYPT) {
			const uint32_t tmp_u32 = cpu_to_le32 (hlen + len + handle->crypto.digest_length);
			const uint64_t tmp_u64 = cpu_to_le64 (handle->crypto.encrypt_cnt);

			EVP_DigestSignInit (handle->crypto.digest_encrypt_ctx, NULL, handle->crypto.digest, NULL, NULL);
			EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u64, sizeof tmp_u64);
			EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, (const unsigned char *) &tmp_u32, sizeof tmp_u32);
			EVP_DigestSignUpdate (handle->crypto.digest_encrypt_ctx, handle->sendbuf + offset, hlen + len);

			size_t siglen = handle->crypto.digest_length;
			EVP_DigestSignFinal (handle->crypto.digest_encrypt_ctx, p, &siglen);
			len += siglen;

			int outlen;
			EVP_EncryptUpdate (handle->crypto.cipher_encrypt_ctx, handle->sendbuf + 5, &outlen, handle->sendbuf + 5, len);
			handle->crypto.encrypt_cnt++;
		}

		len += hlen;

send_again:
		written = send (handle->fd, (void *)(handle->sendbuf + offset), len, 0);

		if ((size_t)written == len) {
			handle->bytes_written += written;
			continue;
		}

#ifdef HAVE_ERRNO_H
		if (written == -1 && errno == EINTR)
			goto send_again;
#endif
		handle->error = MDF_ERR_DISCONNECTED;
		disconnect (handle);
		return 0;
	}

#ifdef TCP_CORK
	i = 0;
	setsockopt (handle->fd, IPPROTO_TCP, TCP_CORK, &i, sizeof (i));
#endif

	should_we_rekey (handle);
	return 1;
}

