/*
    Copyright (C) 2008-2019, Millistream Market Data <support@millistream.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "common.h"
#include "templates.h"
#include "decode.h"
#include "crypt.h"

static void inflate_mmt (const uint8_t *in, uint8_t * const out)
{
	int i;
	for (i = 0; *in != '\0' && i < 14; in++) {
		if (*in >= 'a') {
			int j = *in - 0x60;

			if (j > 14 - i)
				j = 14 - i;

			switch (j) {
				case 14 : out[i++] = '-'; /* FALLTHRU */
				case 13 : out[i++] = '-'; /* FALLTHRU */
				case 12 : out[i++] = '-'; /* FALLTHRU */
				case 11 : out[i++] = '-'; /* FALLTHRU */
				case 10 : out[i++] = '-'; /* FALLTHRU */
				case  9 : out[i++] = '-'; /* FALLTHRU */
				case  8 : out[i++] = '-'; /* FALLTHRU */
				case  7 : out[i++] = '-'; /* FALLTHRU */
				case  6 : out[i++] = '-'; /* FALLTHRU */
				case  5 : out[i++] = '-'; /* FALLTHRU */
				case  4 : out[i++] = '-'; /* FALLTHRU */
				case  3 : out[i++] = '-'; /* FALLTHRU */
				case  2 : out[i++] = '-'; /* FALLTHRU */
				case  1 : out[i++] = '-';
			}
		} else {
			out[i++] = *in;
		}
	}

	/* make sure that we fill the whole MMT string with blanks if needed */
	if (i < 14) {
		switch (14 - i) {
			case 14 : out[i++] = '-'; /* FALLTHRU */
			case 13 : out[i++] = '-'; /* FALLTHRU */
			case 12 : out[i++] = '-'; /* FALLTHRU */
			case 11 : out[i++] = '-'; /* FALLTHRU */
			case 10 : out[i++] = '-'; /* FALLTHRU */
			case  9 : out[i++] = '-'; /* FALLTHRU */
			case  8 : out[i++] = '-'; /* FALLTHRU */
			case  7 : out[i++] = '-'; /* FALLTHRU */
			case  6 : out[i++] = '-'; /* FALLTHRU */
			case  5 : out[i++] = '-'; /* FALLTHRU */
			case  4 : out[i++] = '-'; /* FALLTHRU */
			case  3 : out[i++] = '-'; /* FALLTHRU */
			case  2 : out[i++] = '-'; /* FALLTHRU */
			case  1 : out[i] = '-';
		}
	}

	out[14] = '\0';
}

int zlib_uncompress (uint8_t * restrict src, uint8_t * restrict dst, uLong * restrict slen, uLong * restrict dlen)
{
	z_stream stream;
	int err;

	stream.next_in = (Bytef *) src;
	stream.avail_in = (uInt) *slen;

	/* Check for source > 64K on 16-bit machine: */
	if ((uLong)stream.avail_in != *slen)
		return Z_BUF_ERROR;

	stream.next_out = (Bytef *) dst;
	stream.avail_out = (uInt) *dlen;

	/* Check for destination > 64K on 16-bit machine: */
	if ((uLong)stream.avail_out != *dlen)
		return Z_BUF_ERROR;

	stream.zalloc = (alloc_func) 0;
	stream.zfree = (free_func) 0;

	err = inflateInit (&stream);

	if (err != Z_OK)
		return err;

	err = inflate (&stream, Z_FINISH);

	if (err != Z_STREAM_END) {
		inflateEnd(&stream);

		if (err == Z_NEED_DICT || (err == Z_BUF_ERROR && stream.avail_in == 0))
			return Z_DATA_ERROR;

		return err;
	}

	*slen = stream.total_in;
	*dlen = stream.total_out;
	err = inflateEnd(&stream);

	return err;
}

static int decode_bcd (mdf_t handle)
{
	static const char bcdLUT[14] = "0123456789. -\x00";
	int pos = 0;

	do {
		const uint8_t ch = handle->data[handle->fpos++];

		handle->value[pos++] = bcdLUT[ch >> 4];
		handle->value[pos++] = bcdLUT[ch & 0x0f];

		if ((ch & 0x0f) == 0x0D)
			return 1;

		if ((size_t)(pos + 2) > handle->vlen) {
			void *ptr = realloc (handle->value, handle->vlen + 1024);

			if (ptr == NULL) {
				handle->error = MDF_ERR_NO_MEM;
				return 0;
			}

			handle->value = ptr;
			handle->vlen += 1024;
		}
	} while (handle->fpos < handle->message_end);

	handle->error = MDF_ERR_MSG_OOB;
	return 0;
}

static inline int num_digits (const uint64_t value)
{
	if (value < __UINT64_C(100000)) {
		if (value < __UINT64_C(10)) return 1;
		if (value < __UINT64_C(100)) return 2;
		if (value < __UINT64_C(1000)) return 3;
		if (value < __UINT64_C(10000)) return 4;
		return 5;
	}

	if (value < __UINT64_C(10000000000)) {
		if (value < __UINT64_C(1000000)) return 6;
		if (value < __UINT64_C(10000000)) return 7;
		if (value < __UINT64_C(100000000)) return 8;
		if (value < __UINT64_C(1000000000)) return 9;
		return 10;
	}

	if (value < __UINT64_C(1000000000000000)) {
		if (value < __UINT64_C(100000000000)) return 11;
		if (value < __UINT64_C(1000000000000)) return 12;
		if (value < __UINT64_C(10000000000000)) return 13;
		if (value < __UINT64_C(100000000000000)) return 14;
		return 15;
	}

	if (value < __UINT64_C(10000000000000000)) return 16;
	if (value < __UINT64_C(100000000000000000)) return 17;
	if (value < __UINT64_C(1000000000000000000)) return 18;
	if (value < __UINT64_C(10000000000000000000)) return 19;

	return 20;
}

static const char itoaLUT[201] =
	"0001020304050607080910111213141516171819"
	"2021222324252627282930313233343536373839"
	"4041424344454647484950515253545556575859"
	"6061626364656667686970717273747576777879"
	"8081828384858687888990919293949596979899";

/* u8 to string, does however support 999-0 and not 255-0 */
static inline int u8toa (const unsigned int value, char * const string)
{
	if (value >= 100) {
		memcpy (string, itoaLUT + ((value / 10) * 2), 2);
		string[2] = '0' + value % 10;
		string[3] = '\0';
		return 3;
	} else if (value >= 10) {
		memcpy (string, itoaLUT + (value * 2), 2);
		string[2] = '\0';
		return 2;
	}

	string[0] = '0' + value;
	string[1] = '\0';
	return 1;
}

/* The u32toa() and u64toa() functions modified from https://github.com/amdn/itoa_ljust/
 * Copyright (c) 2016 Arturo Martin-de-Nicolas
 * arturomdn@gmail.com
 *
 * Released under the BSD 3-Clause License, see Google's original copyright
 * and license below.
 *
 * Protocol Buffers - Google's data interchange format
 * Copyright 2008 Google Inc.  All rights reserved.
 * https://developers.google.com/protocol-buffers/
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

static inline int u32toa (uint32_t value, char *string)
{
	const char *old = string;
	int d,n;

	if (value >= 100000000) {
		if (value < 100000000 * 10) {
			d = value / 100000000;
			*string++ = '0' + d;
			n = 9;
		} else {
			n = 10;
		}
	} else if (value < 100) {
		if (value < 10) {
			*string++ = '0' + value;
			n = 1;
		} else {
			n = 2;
		}
	} else if (value < 10000) {
		if (value < 100 * 10) {
			d = value / 100;
			*string++ = '0' + d;
			n = 3;
		} else {
			n = 4;
		}
	} else if (value < 1000000) {
		if (value < 10000 * 10) {
			d = value / 10000;
			*string++ = '0' + d;
			n = 5;
		} else {
			n = 6;
		}
	} else if (value < 1000000 * 10) {
		d = value / 1000000;
		*string++ = '0' + d;
		n = 7;
	} else {
		n = 8;
	}

	switch (n) {
		case 10: d  = value / 100000000; memcpy (string, itoaLUT + d * 2, 2); string += 2; /* FALLTHRU */
		case  9: value -= d * 100000000; /* FALLTHRU */
		case  8: d  = value /   1000000; memcpy (string, itoaLUT + d * 2, 2); string += 2; /* FALLTHRU */
		case  7: value -= d *   1000000; /* FALLTHRU */
		case  6: d  = value /     10000; memcpy (string, itoaLUT + d * 2, 2); string += 2; /* FALLTHRU */
		case  5: value -= d *     10000; /* FALLTHRU */
		case  4: d  = value /       100; memcpy (string, itoaLUT + d * 2, 2); string += 2; /* FALLTHRU */
		case  3: value -= d *       100; /* FALLTHRU */
		case  2: memcpy (string, itoaLUT + value * 2, 2); string += 2; /* FALLTHRU */
	}

	*string = '\0';

	return string - old;
}

static inline int u64toa (const uint64_t value, char *string)
{
	uint32_t lower = (uint32_t) (value);

	if (lower == value)
		return u32toa (lower, string);

	uint64_t upper = value / 1000000000;

	int len = u64toa (upper, string);

	lower = value - (upper * 1000000000);
	int d = lower / 100000000;

	string += len;
	*string++ = '0' + d;

	lower -= d * 100000000;
	d  = lower / 1000000;
	memcpy (string, itoaLUT + d * 2, 2); string += 2;

	lower -= d * 1000000;
	d  = lower / 10000;
	memcpy (string, itoaLUT + d * 2, 2); string += 2;

	lower -= d * 10000;
	d  = lower / 100;
	memcpy (string, itoaLUT + d * 2, 2); string += 2;

	lower -= d * 100;
	memcpy (string, itoaLUT + lower * 2, 2); string += 2;
	*string = '\0';

	return len + 9;
}

static int parse_intarray (mdf_t handle, char prefix)
{
	uint64_t num = 0;
	size_t len;
	int pos = 0;

	do {
		if (handle->fpos >= handle->message_end)
			return 0;

		num *= 128;
		num += (handle->data[handle->fpos] & 127) + 1;
	} while ((handle->data[handle->fpos++] & 128) == 0);

	/* safety limit */
	if (num > 1000000) {
		handle->error = MDF_ERR_MSG_OOB;
		return 0;
	}

	len = (size_t) (num * 21 + 2);

	if (len > handle->vlen) {
		free (handle->value);
		handle->value = malloc (len);

		if (handle->value == NULL) {
			handle->vlen = 1024;
			handle->value = malloc (handle->vlen);
			handle->error = MDF_ERR_NO_MEM;
			return 0;
		}

		handle->vlen = len;
	}

	if (prefix != ' ')
		handle->value[pos++] = (uint8_t) prefix;

	for (; num != 0; --num) {
		uint64_t v = 0;

		do {
			if (handle->fpos >= handle->message_end)
				return 0;

			v *= 128;
			v += (handle->data[handle->fpos] & 127) + 1;
		} while ((handle->data[handle->fpos++] & 128) == 0);

		pos += u64toa (v - 1, (char *)(handle->value + pos));
		handle->value[pos++] = ' ';
	}

	/* remove the last space */
	handle->value[pos - 1] = '\0';
	return 1;
}

static void parse_msg_header (const uint8_t * const data, size_t *pos, uint64_t * const insref, uint16_t * const mref)
{
	const size_t f_pos = *pos;

	/* the message reference is either 8 or 16 bits */
	if (data[f_pos] & 128) {
		*mref = le16_to_cpu (*((const uint16_t *) (data + f_pos + 1)));
		*pos += 3;
	} else {
		*mref = data[f_pos + 1];
		*pos += 2;
	}

	/* the instrument reference is 8, 16, 32 or 64 bits */
	switch (data[f_pos] & 96) {
		case 0 :
			*insref = data[*pos] + ((uint64_t)(data[f_pos] & 31) << 8);
			(*pos)++;
			break;

		case 32 :
			*insref = le16_to_cpu (*((const uint16_t *)(data + *pos))) + ((uint64_t)(data[f_pos] & 31) << 16) + 8192;
			*pos += 2;
			break;

		case 64 :
			*insref = le32_to_cpu (*((const uint32_t *)(data + *pos))) + ((uint64_t)(data[f_pos] & 31) << 32) + 2105344;
			*pos += 4;
			break;

		default :
			*insref = le64_to_cpu (*((const uint64_t *)(data + *pos)));
			*pos += 8;
			break;
	}
}

LIBMDF_API int mdf_get_next_message (mdf_t handle, int *message_reference, int *message_class, uint64_t *instrument_reference)
{
	if (handle == NULL)
		return 0;

	handle->error = MDF_ERR_NO_ERROR;

try_again:
	if (handle->fstatus == 0) {
		int pos = 0;
		size_t tlen = 0;

		handle->data_pos += handle->message_len;

		if (handle->data_pos >= handle->data_used) {
			handle->data_pos = 0;
			handle->data_used = 0;
			handle->fstatus = 2;
			return 0;
		}

		handle->data[handle->message_end] = handle->oldval; /* we zeroed this byte before */

		do {
			tlen *= 128;
			tlen += (handle->data[handle->data_pos + pos] & 127) + 1;
		} while ((handle->data[handle->data_pos + pos++] & 128) == 0 && handle->data_pos + pos < handle->data_used);

		if (pos > 5) {
			mdf_disconnect (handle);
			handle->error = MDF_ERR_MSG_OOB;
			return 0;
		}

		/* we have not yet received a full header */
		if (pos == 0 || (handle->data[handle->data_pos + pos - 1] & 128) == 0) {
			if (handle->data_pos != 0) {
				handle->data_used -= handle->data_pos;
				memmove (handle->data, handle->data + handle->data_pos, handle->data_used);
				handle->data_pos = 0;
			}

			handle->fstatus = 2;
			return 0;
		}

		handle->message_len = tlen + pos;
		handle->message_end = handle->data_pos + handle->message_len;
		handle->pos = pos;
		handle->fpos = handle->data_pos + pos;

		if ((handle->crypto.cstate & CSTATE_DECRYPT) == CSTATE_DECRYPT)
			handle->message_len += handle->crypto.digest_length;

		if (handle->data_pos + handle->message_len + 12 > handle->data_size) {
			void *tmp = realloc (handle->data, handle->data_pos + handle->message_len + 12);

			if (tmp == NULL) {
				mdf_disconnect (handle);
				handle->error = MDF_ERR_NO_MEM;
				return 0;
			}

			handle->data = (uint8_t *) tmp;
			handle->data_size = handle->data_pos + handle->message_len + 12;
		}

		if (handle->message_len + handle->data_pos > handle->data_used) {
			if (handle->data_pos != 0) {
				handle->data_used -= handle->data_pos;
				memmove (handle->data, handle->data + handle->data_pos, handle->data_used);
				handle->data_pos = 0;
			}

			handle->fstatus = 2;
			return 0;
		}
	} else if (handle->fstatus == 2) {
		return 0;
	}

	if ((handle->crypto.cstate & CSTATE_DECRYPT) == CSTATE_DECRYPT) {
		unsigned char md[EVP_MAX_MD_SIZE];
		const uint64_t decrypt_cnt = cpu_to_le64 (handle->crypto.decrypt_cnt);
		const uint32_t tmp_len = cpu_to_le32 (handle->message_len);

		EVP_DigestSignInit (handle->crypto.digest_decrypt_ctx, NULL, handle->crypto.digest, NULL, NULL);
		EVP_DigestSignUpdate (handle->crypto.digest_decrypt_ctx, (const unsigned char *) &decrypt_cnt, sizeof decrypt_cnt);
		EVP_DigestSignUpdate (handle->crypto.digest_decrypt_ctx, (const unsigned char *) &tmp_len, sizeof tmp_len);

		int outlen;
		EVP_DecryptUpdate (handle->crypto.cipher_decrypt_ctx, handle->data + handle->data_pos + handle->pos, &outlen, handle->data + handle->data_pos + handle->pos, handle->message_len - handle->pos);
		handle->crypto.decrypt_cnt++;

		EVP_DigestSignUpdate (handle->crypto.digest_decrypt_ctx, handle->data + handle->data_pos, handle->message_len - handle->crypto.digest_length);

		size_t siglen = sizeof md;
		EVP_DigestSignFinal (handle->crypto.digest_decrypt_ctx, md, &siglen);

		if (memcmp (handle->data + handle->data_pos + handle->message_len - handle->crypto.digest_length, md, handle->crypto.digest_length) != 0) {
			mdf_disconnect (handle);
			handle->error = MDF_ERR_AUTHFAIL;
			return 0;
		}
	}

	/* so we can safely use strlen() & co */
	handle->oldval = handle->data[handle->message_end];
	handle->data[handle->message_end] = '\0';

	handle->pos = handle->fpos;

	uint64_t insref;
	uint16_t mref;
	parse_msg_header (handle->data, &handle->pos, &insref, &mref);

	handle->fpos = handle->pos + 1;
	handle->in_pmap_pos = 128;
	handle->field_pmap_pos = 0;
	handle->fstatus = 0;

	if (mref == 0) { /* hardcoded message */
		if (insref == 1) { /* heartbeat request */
			if (mdf_int_send_heartbeat (handle) == -1)
				return 0;
		} else if (insref == 2) { /* heartbeat response */
			if (handle->message_end - handle->pos >= 8) {
				int64_t heartbeat = le64_to_cpu (*((int64_t *)(handle->data + handle->pos)));
				handle->timediff = heartbeat - time (NULL);
			}
		} else if (insref == 5) { /* new keys from server */
			mdf_int_handle_server_rekey (&handle->crypto, handle->data + handle->pos, handle->message_end - handle->pos);
		}

		goto try_again;
	} else if (mref == MDF_M_MESSAGESREFERENCE + 1) {
		handle->template_version = insref;
		templates_decode (handle->data + handle->pos, handle->data + (handle->message_end - handle->pos), &handle->templates, &handle->templates_num);
	} else if (mref == MDF_M_LOGONGREETING + 1 && handle->no_encryption == 1) {
		mdf_int_send_rekey (handle);
	}

	/* mref 0 have no template */
	mref--;

	/* since this check is done after we may have parsed the
	 * msg header we must make sure that the data buffer always
	 * is at least 12 bytes larger than the actual tlen! */
	if (handle->fpos > handle->message_end) {
		handle->error = MDF_ERR_MSG_OOB;
		return 0;
	}

	*message_reference = (int) mref;
	*instrument_reference = insref;

	if (mref >= handle->templates_num) {
		*message_class = MDF_MC_UNDEF;
		handle->template = NULL;
	} else {
		handle->template = &handle->templates[mref];
		*message_class = handle->template->mclass;
	}

	return 1;
}

LIBMDF_API int mdf_get_next_field (mdf_t handle, uint32_t *tag, char **value)
{
	uint8_t field_code;

	if (handle == NULL)
		return 0;

	/* someone told us to skip all fields in this message or we where not called
	 * after a call to mdf_get_next_message() that returned true */
	if (handle->fstatus != 0)
		return 0;

	if (handle->fpos >= handle->message_end)
		return 0;

	/* unknown template */
	if (handle->template == NULL || handle->template->tags == NULL) {
		handle->error = MDF_ERR_UNKNOWN_TEMPLATE;
		return 0;
	}

	/* let's find the first used bit in the pmap */
	while ((handle->data[handle->pos] & handle->in_pmap_pos) == 0) {
		handle->in_pmap_pos /= 2;
		handle->field_pmap_pos++;

		/* no more bits in this pmap, go to next pmap */
		if (handle->in_pmap_pos == 0) {
			handle->pos = handle->fpos;
			handle->fpos++;
			handle->in_pmap_pos = 128;

			if (handle->fpos >= handle->message_end)
				return 0;
		}
	}

	/* our template is out-of-date or we have gotten erreneous input, bail out */
	if (handle->field_pmap_pos >= handle->template->tags_num) {
		handle->error = MDF_ERR_TEMPLATE_OOB;
		return 0;
	}

	field_code = handle->data[handle->fpos];

	handle->fpos++;

	/* Tagged fields will always precede the dynamic (aka template) fields 
	 * in the message. And since they do not have a slot in the pmap we have 
	 * to compensate [in_pmap_pos] for the division that happens at the end 
	 * of this function! */
	if (field_code < 251 || field_code > 253) { /* Dynamic field */
		*tag = handle->template->tags[handle->field_pmap_pos];
	} else if (field_code == 251) { /* Tagged field uint8 */
		*tag = handle->data[handle->fpos];
		field_code = handle->data[handle->fpos + 1];
		handle->fpos += 2;
		handle->in_pmap_pos *= 2;
		handle->field_pmap_pos--;
	} else if (field_code == 252) { /* Tagged field uint16 */
		*tag = le16_to_cpu (*((uint16_t *)(handle->data + handle->fpos)));
		field_code = handle->data[handle->fpos + 2];
		handle->fpos += 3;
		handle->in_pmap_pos *= 2;
		handle->field_pmap_pos--;
	} else if (field_code == 253) { /* Tagged field uint32 */
		*tag = le32_to_cpu (*((uint32_t *)(handle->data + handle->fpos)));
		field_code = handle->data[handle->fpos + 4];
		handle->fpos += 5;
		handle->in_pmap_pos *= 2;
		handle->field_pmap_pos--;
	}

	*value = (char *) handle->value;

	/* field_codes that are currently not in use: 224 - 239, 243, 245, 254 */

	if (field_code <= 127) {
		if (field_code == 0) {
			u8toa (((unsigned int) handle->data[handle->fpos++]) + 28, *value);
		} else if (field_code > 99) {
			u8toa (field_code - 100, *value);
		} else {
			int pos = u8toa ((unsigned int) handle->data[handle->fpos++], *value);
			(*value)[pos++] = '.';

			if (field_code < 10)
				(*value)[pos++] = '0';

			u8toa (field_code, *value + pos);
		}
	} else if (field_code <= 143) { /* .xxxx */
		const uint32_t v = le16_to_cpu (*((uint16_t *)(handle->data + handle->fpos))) + ((field_code & 15) << 16) + 1;
		const uint32_t r = v / 10000;
		const uint32_t m = v - (r * 10000);
		int pos = u8toa (r, *value);

		if (m != 0) {
			(*value)[pos++] = '.';

			if (m < 1000) (*value)[pos++] = '0';
			if (m < 100) (*value)[pos++] = '0';
			if (m < 10) (*value)[pos++] = '0';

			u32toa (m, *value + pos);
		}

		handle->fpos += 2;
	} else if (field_code <= 159) { /* 10741.xx */
		const uint32_t v = le16_to_cpu (*((uint16_t *)(handle->data + handle->fpos))) + ((field_code & 15) << 16) + 25601;
		const uint32_t r = v / 100;
		const uint32_t m = v - (r * 100);
		int pos = u32toa (r, *value);

		if (m != 0) {
			(*value)[pos++] = '.';

			if (m < 10) (*value)[pos++] = '0';

			u8toa (m, *value + pos);
		}

		handle->fpos += 2;
	} else if (field_code <= 175) { /* Positive integer with 1-16 zeroes */
		uint64_t v = 0;
		int len, num;

		do {
			if (handle->fpos >= handle->message_end) {
				handle->error = MDF_ERR_MSG_OOB;
				return 0;
			}

			v *= 128;
			v += (handle->data[handle->fpos] & 127) + 1;
		} while ((handle->data[handle->fpos++] & 128) == 0);

		len = u64toa (v, *value);
		num = (field_code & 15) + 1;

		memset (*value + len, '0', num);
		(*value)[len + num] = '\0';
	} else if (field_code <= 191) { /* Positive number with 0-15 decimals */
		uint64_t v = 0;
		int decimals;

		do {
			if (handle->fpos >= handle->message_end) {
				handle->error = MDF_ERR_MSG_OOB;
				return 0;
			}

			v *= 128;
			v += (handle->data[handle->fpos] & 127) + 1;
		} while ((handle->data[handle->fpos++] & 128) == 0);

		decimals = field_code & 15;

		if (decimals != 0) {
			const int digits = num_digits (v);

			if (decimals >= digits) {
				const int diff = decimals - digits;

				memcpy (*value, "0.", 2);
				memset (*value + 2, '0', diff);
				u64toa (v, *value + diff + 2);
			} else {
				const int diff = digits - decimals;

				u64toa (v, *value);
				memmove (*value + diff + 1, *value + diff, decimals + 1);
				(*value)[diff] = '.';
			}
		} else {
			u64toa (v, *value);
		}
	} else if (field_code <= 207) { /* Negative number with 0-15 decimals */
		uint64_t v = 0;
		int decimals;

		do {
			if (handle->fpos >= handle->message_end) {
				handle->error = MDF_ERR_MSG_OOB;
				return 0;
			}

			v *= 128;
			v += (handle->data[handle->fpos] & 127) + 1;
		} while ((handle->data[handle->fpos++] & 128) == 0);

		decimals = field_code & 15;

		if (decimals != 0) {
			const int digits = num_digits (v);

			if (decimals >= digits) {
				const int diff = decimals - digits;

				memcpy (*value, "-0.", 3);
				memset (*value + 3, '0', diff);
				u64toa (v, *value + diff + 3);
			} else {
				const int diff = digits - decimals + 1;

				(*value)[0] = '-';
				u64toa (v, *value + 1);
				memmove (*value + diff + 1, *value + diff, decimals + 1);
				(*value)[diff] = '.';
			}
		} else {
			(*value)[0] = '-';
			u64toa (v, *value + 1);
		}
	} else if (field_code <= 223) { /* Date */
		int years, months;
		int days = le16_to_cpu (*((uint16_t *)(handle->data + handle->fpos))) + ((field_code & 15) << 16);
		int pos;

		years = days / 448; days -= years * 448;
		months = days / 32; days -= months * 32;

		if (days == 0) {
			if (months == 0) {
				u32toa (years, *value);
			} else {
				pos = u32toa (years, *value);
				(*value)[pos++] = '-';

				if (months < 10)
					(*value)[pos++] = '0';

				u8toa (months, *value + pos);
			}
		} else if (months == 0) {
			pos = u32toa (years, *value);
			(*value)[pos++] = '-';

			switch (days) {
				case 1 : memcpy (*value + pos, "H1", 3); break;
				case 2 : memcpy (*value + pos, "H2", 3); break;
				case 3 : memcpy (*value + pos, "T1", 3); break;
				case 4 : memcpy (*value + pos, "T2", 3); break;
				case 5 : memcpy (*value + pos, "T3", 3); break;
				case 6 : memcpy (*value + pos, "Q1", 3); break;
				case 7 : memcpy (*value + pos, "Q2", 3); break;
				case 8 : memcpy (*value + pos, "Q3", 3); break;
				case 9 : memcpy (*value + pos, "Q4", 3); break;
				default: (*value)[pos++] = 'W'; u8toa (days - 9, *value + pos); break;
			}
		} else if (months == 13) {
			pos = u32toa (years, *value);
			(*value)[pos++] = '-';
			(*value)[pos++] = 'W';
			u8toa (days + 22, *value + pos);
		} else {
			pos = u32toa (years, *value);

			(*value)[pos++] = '-';

			if (months < 10) (*value)[pos++] = '0';
			pos += u8toa (months, *value + pos);

			(*value)[pos++] = '-';

			if (days < 10) (*value)[pos++] = '0';
			u8toa (days, *value + pos);
		}

		handle->fpos += 2;
	} else {
		switch (field_code) {
			case 240 : /* NULL */
				*value = NULL;
				break;

			case 241 : { /* Uncompressed String */
				const size_t len = strlen ((char *) (handle->data + handle->fpos)) + 1;
				*value = (char *)(handle->data + handle->fpos);

				handle->fpos += len;

				if (*tag == MDF_F_MMT) {
					inflate_mmt ((uint8_t *)*value, handle->value);
					*value = (char *) handle->value;
				}

				break;
			}

			case 242 : { /* Compressed String */
				uLong slen;
				uLong dlen = 0;

				do {
					if (handle->fpos >= handle->message_end) {
						handle->error = MDF_ERR_MSG_OOB;
						return 0;
					}

					dlen *= 128;
					dlen += (handle->data[handle->fpos] & 127) + 1;
				} while ((handle->data[handle->fpos++] & 128) == 0);

				if (dlen + 1 > handle->vlen) {
					free (handle->value);
					handle->value = malloc (dlen + 1);

					if (handle->value == NULL) {
						handle->vlen = 1024;
						handle->value = malloc (handle->vlen);
						handle->error = MDF_ERR_NO_MEM;
						return 0;
					}

					handle->vlen = dlen + 1;
					*value = (char *) handle->value;
				}

				slen = handle->message_end - handle->fpos;

				if (zlib_uncompress (handle->data + handle->fpos, handle->value, &slen, &dlen) != Z_OK) {
					handle->error = MDF_ERR_MSG_OOB;
					return 0;
				}

				handle->fpos += (uint32_t) slen;
				handle->value[dlen] = '\0';
				break;
			}

			case 244 : { /* Time Value Type #3 */
				int pos = 0;

				if ((handle->data[handle->fpos + 3] & 128) == 128) { /* nano seconds */
					union {
						uint64_t val;
						uint8_t arr[8];
					} nsec;

					nsec.arr[6] = 0;
					nsec.arr[7] = 0;
					nsec.arr[4] = handle->data[handle->fpos++];
					nsec.arr[3] = handle->data[handle->fpos++];
					nsec.arr[2] = handle->data[handle->fpos++];
					nsec.arr[5] = handle->data[handle->fpos++] & 127;
					nsec.arr[1] = handle->data[handle->fpos++];
					nsec.arr[0] = handle->data[handle->fpos++];
					nsec.val = le64_to_cpu (nsec.val);

					const int hours = nsec.val / __UINT64_C(3600000000000); nsec.val -= hours * __UINT64_C(3600000000000);
					const int minutes = nsec.val / __UINT64_C(60000000000); nsec.val -= minutes * __UINT64_C(60000000000);
					const int seconds = nsec.val / __UINT64_C(1000000000); nsec.val -= seconds * __UINT64_C(1000000000);

					if (hours < 10) (*value)[pos++] = '0';
					pos += u8toa (hours, *value + pos);
					(*value)[pos++] = ':';

					if (minutes < 10) (*value)[pos++] = '0';
					pos += u8toa (minutes, *value + pos);
					(*value)[pos++] = ':';

					if (seconds < 10) (*value)[pos++] = '0';
					pos += u8toa (seconds, *value + pos);
					(*value)[pos++] = '.';

					if (nsec.val < __UINT64_C(100000000)) (*value)[pos++] = '0';
					if (nsec.val < 10000000) (*value)[pos++] = '0';
					if (nsec.val < 1000000) (*value)[pos++] = '0';
					if (nsec.val < 100000) (*value)[pos++] = '0';
					if (nsec.val < 10000) (*value)[pos++] = '0';
					if (nsec.val < 1000) (*value)[pos++] = '0';
					if (nsec.val < 100) (*value)[pos++] = '0';
					if (nsec.val < 10) (*value)[pos++] = '0';
					u32toa (nsec.val, *value + pos);
				} else { /* milliseconds */
					uint32_t msec = le32_to_cpu (*((uint32_t *)(handle->data + handle->fpos)));
					handle->fpos += 4;

					const int hours = msec / 3600000; msec -= hours * 3600000;
					const int minutes = msec / 60000; msec -= minutes * 60000;
					const int seconds = msec / 1000; msec -= seconds * 1000;

					if (hours < 10) (*value)[pos++] = '0';
					pos += u8toa (hours, *value + pos);
					(*value)[pos++] = ':';

					if (minutes < 10) (*value)[pos++] = '0';
					pos += u8toa (minutes, *value + pos);
					(*value)[pos++] = ':';

					if (seconds < 10) (*value)[pos++] = '0';
					pos += u8toa (seconds, *value + pos);
					(*value)[pos++] = '.';

					if (msec < 100) (*value)[pos++] = '0';
					if (msec < 10) (*value)[pos++] = '0';
					u8toa (msec, *value + pos);
				}

				break;
			}

			case 246 : { /* Time Value Type #1 */
				int hours, minutes;
				int seconds = le16_to_cpu (*((uint16_t *)(handle->data + handle->fpos)));
				int pos = 0;

				hours = seconds / 3600; seconds -= hours * 3600;
				minutes = seconds / 60; seconds -= minutes * 60;

				if (hours < 10) (*value)[pos++] = '0';
				pos += u8toa (hours, *value + pos);
				(*value)[pos++] = ':';

				if (minutes < 10) (*value)[pos++] = '0';
				pos += u8toa (minutes, *value + pos);
				(*value)[pos++] = ':';

				if (seconds < 10) (*value)[pos++] = '0';
				u8toa (seconds, *value + pos);

				handle->fpos += 2;
				break;
			}

			case 247 : { /* Time Value Type #2 */
				int hours, minutes;
				int seconds = le16_to_cpu (*((uint16_t *)(handle->data + handle->fpos))) + 65535;
				int pos = 0;
				
				hours = seconds / 3600; seconds -= hours * 3600;
				minutes = seconds / 60; seconds -= minutes * 60;

				if (hours < 10) (*value)[pos++] = '0';
				pos += u8toa (hours, *value + pos);
				(*value)[pos++] = ':';

				if (minutes < 10) (*value)[pos++] = '0';
				pos += u8toa (minutes, *value + pos);
				(*value)[pos++] = ':';

				if (seconds < 10) (*value)[pos++] = '0';
				u8toa (seconds, *value + pos);

				handle->fpos += 2;
				break;
			}

			case 248 : /* IntArray(+) */
				if (!parse_intarray (handle, '+'))
					return 0;

				*value = (char *) handle->value; /* since handle->value could be reallocated */
				break;

			case 249 : /* IntArray(-) */
				if (!parse_intarray (handle, '-'))
					return 0;

				*value = (char *) handle->value; /* since handle->value could be reallocated */
				break;

			case 250 : /* IntArray(=) */
				if (!parse_intarray (handle, *tag != MDF_F_INSREFLIST ? '=' : ' '))
					return 0;

				*value = (char *) handle->value; /* since handle->value could be reallocated */
				break;

			case 255 : /* Packed BCD */
				if (!decode_bcd (handle))
					return 0;

				break;
		}
	}

	/* so we don't loop forever on the same pmap bit on every call to this function */
	handle->in_pmap_pos /= 2;
	handle->field_pmap_pos++;

	/* if no more bits in this pmap, go to next pmap */
	if (handle->in_pmap_pos == 0) {
		handle->pos = handle->fpos;
		handle->fpos++;
		handle->in_pmap_pos = 128;
	}

	return 1;
}

