/*
    Copyright (C) 2008-2019, Millistream Market Data <support@millistream.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "common.h"
#include "decode.h"
#include "templates.h"

static int compare_tags (const void *A, const void *B)
{
	return (((const struct msg_field *)A)->tag - ((const struct msg_field *)B)->tag);
}

static void templates_rebalance (struct msg_template * const templates, const int templates_num)
{
	int i, j;
	for (i = 0; i < templates_num; i++) {
		free (templates[i].fields);

		if (templates[i].tags_num == 0) {
			templates[i].fields = NULL;
			continue;
		}

		templates[i].fields = (struct msg_field *) malloc ((sizeof *templates[i].fields) * templates[i].tags_num);

		for (j = 0; j < templates[i].tags_num; j++) {
			templates[i].fields[j].pos = j;
			templates[i].fields[j].tag = templates[i].tags[j];
		}

		qsort (templates[i].fields, templates[i].tags_num, sizeof *templates[i].fields, compare_tags);
	}
}

static void templates_update (struct msg_template ** const templates, int * const templates_num, const int mref, const char *p)
{
	const char *e;
	uint64_t mclass;
	int tags_num = 0;
	int i;
	uint32_t *tags;

	if (mref > 65535)
		return;

	if (mref >= *templates_num) {
		*templates = realloc (*templates, (sizeof **templates) * (mref + 1));
		memset (&((*templates)[*templates_num]), 0, (sizeof **templates) * ((mref - *templates_num) + 1));
		*templates_num = mref + 1;
	}

	p = strchr (p, '|');

	if (p == NULL)
		return;

	p = strchr (p + 1, '|');

	if (p == NULL)
		return;

	p++;

	mclass = strtoull (p, NULL, 10);

	p = strchr (p, '|');

	if (p == NULL)
		return;

	p++;

	for (e = p; *e != '\0' && *e != '|'; e++) {
		if (*e == ' ') {
			tags_num++;

			while (e[1] == ' ')
				e++;
		}
	}

	if (*p != '\0' && *p != '|')
		tags_num++;

	if ((*templates)[mref].tags != NULL) {
		free ((*templates)[mref].tags);
		(*templates)[mref].tags = NULL;
	}

	(*templates)[mref].mclass = mclass;
	(*templates)[mref].tags_num = tags_num;

	if (tags_num != 0)
		(*templates)[mref].tags = (uint32_t *) malloc (sizeof (uint32_t) * tags_num);

	tags = (*templates)[mref].tags;

	for (i = 0; i < tags_num; i++) {
		while (*p == ' ')
			p++;

		tags[i] = strtoul (p, NULL, 10);

		p = strchr (p, ' ');
	}
}

void templates_decode (uint8_t *data, const uint8_t * const end, struct msg_template ** const templates, int * const templates_num)
{
	/* the templates message will always be a fully tagged message consting of only
	 * plain strings and compressed strings. */

	data++; /* skip initial pmap */

	do {
		uint8_t field_code = *data++;
		uint32_t tag;

		if (field_code == 251) { /* Tagged field uint8 */
			tag = *data++;
		} else if (field_code == 252) { /* Tagged field uint16 */
			tag = *((const uint16_t *)(data));
			data += 2;
		} else if (field_code == 253) { /* Tagged field uint32 */
			tag = *((const uint32_t *)(data));
			data += 4;
		}

		field_code = *data++;

		if (field_code == 241) { /* string */
			const size_t slen = strlen ((const char *) (data)) + 1;

			templates_update (templates, templates_num, tag, (char *)data);

			data += slen;
		} else if (field_code == 242) { /* compressed string */
			uLong dlen = 0;

			do {
				if (data >= end)
					break;

				dlen *= 128;
				dlen += (*data & 127) + 1;
			} while ((*data++ & 128) == 0);

			uLong slen = end - data;

			uint8_t *p = malloc (dlen + 1);

			if (zlib_uncompress (data, p, &slen, &dlen) != Z_OK) {
				free (p);
				break;
			}

			p[dlen] = '\0';

			templates_update (templates, templates_num, tag, (char *)p);

			free (p);

			data += slen;
		}
	} while (data < end);

	templates_rebalance (*templates, *templates_num);
}

void templates_initialize (struct msg_template ** const templates, int * const templates_num)
{
	/* setup version #0 of the message templates, this is enough to 
	 * be able to receive the message templates message, to create
	 * a logon message and to receive a possible logoff message */
	*templates_num = 3;
	*templates = calloc (*templates_num, sizeof **templates);

	(*templates)[MDF_M_MESSAGESREFERENCE].mclass = MDF_MC_UNDEF;
	(*templates)[MDF_M_MESSAGESREFERENCE].tags_num = 1;
	(*templates)[MDF_M_MESSAGESREFERENCE].tags = (uint32_t *) malloc (sizeof (uint32_t) * (*templates)[MDF_M_MESSAGESREFERENCE].tags_num);
	(*templates)[MDF_M_MESSAGESREFERENCE].tags[0] = 0xffffffff;

	(*templates)[MDF_M_LOGON].mclass = MDF_MC_UNDEF;
	(*templates)[MDF_M_LOGON].tags_num = 3;
	(*templates)[MDF_M_LOGON].tags = (uint32_t *) malloc (sizeof (uint32_t) * (*templates)[MDF_M_LOGON].tags_num);
	(*templates)[MDF_M_LOGON].tags[0] = MDF_F_USERNAME;
	(*templates)[MDF_M_LOGON].tags[1] = MDF_F_PASSWORD;
	(*templates)[MDF_M_LOGON].tags[2] = MDF_F_EXTRACREDENTIAL;

	(*templates)[MDF_M_LOGOFF].mclass = MDF_MC_UNDEF;
	(*templates)[MDF_M_LOGOFF].tags_num = 1;
	(*templates)[MDF_M_LOGOFF].tags = (uint32_t *) malloc (sizeof (uint32_t) * (*templates)[MDF_M_LOGOFF].tags_num);
	(*templates)[MDF_M_LOGOFF].tags[0] = MDF_F_LOGOFFREASON;

	templates_rebalance (*templates, *templates_num);
}

